# -*- encoding: utf-8 -*-
import unittest

import numpy as np
from autosklearn.metrics import balanced_accuracy, pac_score


def copy_and_preprocess_arrays(solution, prediction):
    solution = solution.copy()
    prediction = prediction.copy()
    return solution, prediction


class BalancedAccurayTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def _test_cases(self, cases):
        for case in cases:
            testname, sol, pred, result = case

            pred = pred.astype(np.float32)
            sol, pred = copy_and_preprocess_arrays(sol, pred)
            bac = balanced_accuracy(sol, pred)
            self.assertAlmostEqual(bac, result, msg=testname)

    def test_binary_balanced(self):
        cases = []
        sol = np.array([0, 0, 1, 1])
        pred = np.array([0, 0, 1, 1])

        cases.append(('perfect', sol, pred, 1.0))
        cases.append(('completely wrong', sol, 1 - pred, 0.0))

        pred = np.array([0, 1, 1, 1])
        cases.append(('partially correct 1', sol, pred, 0.75))

        pred = np.array([0, 1, 0, 1])
        cases.append(('partially correct 2', sol, pred, 0.5))

        pred = np.array([0, 1, 0, 0])
        cases.append(('partially correct 3', sol, pred, 0.25))

        self._test_cases(cases)

    def test_binary_imbalanced(self):
        cases = []
        sol = np.array([0, 1, 1])
        pred = np.array([0, 1, 1])

        cases.append(('perfect', sol, pred, 1.0))
        cases.append(('completely wrong', sol, 1 - pred, 0.0))

        pred = np.array([0, 0, 0])
        cases.append(('one class correct', sol, pred, 0.5))

        pred = np.array([0, 1, 0])
        cases.append(('one class correct, one partially correct',
                      sol, pred, 0.75))

        pred = np.array([1, 0, 1])
        cases.append(('one class partially correct', sol, pred, 0.25))

        self._test_cases(cases)

    def test_multiclass_balanced(self):
        cases = []
        sol = np.array([0, 0, 1, 1, 2, 2])
        pred = np.array([0, 0, 1, 1, 2, 2])

        cases.append(('perfect', sol, pred, 1.0))

        pred = np.array([1, 1, 2, 2, 0, 0])
        cases.append(('completely wrong', sol, pred, 0.0))

        pred = np.array([0, 0, 0, 0, 0, 0])
        cases.append(('one class correct', sol, pred, 0.33333333))

        pred = np.array([0, 0, 1, 1, 0, 0])
        cases.append(('two classes correct', sol, pred, 0.66666666))

        pred = np.array([0, 0, 1, 0, 2, 0])
        cases.append(('one class correct, two partially correct', sol, pred, 0.66666666))

        pred = np.array([0, 1, 1, 2, 2, 0])
        cases.append(('all partially correct', sol, pred, 0.5))
        self._test_cases(cases)

    def test_multiclass_imbalanced(self):
        cases = []
        sol = np.array([0, 1, 2, 0])
        pred = np.array([0, 1, 2, 0])

        cases.append(('all classes perfect', sol, pred, 1.0))

        pred = np.array([1, 2, 0, 1])
        cases.append(('all classes wrong', sol, pred, 0.0))

        pred = np.array([0, 0, 0, 0])
        cases.append(('one class correct', sol, pred, 0.33333333))

        pred = np.array([2, 0, 0, 0])
        cases.append(('one class half-correct', sol, pred, 0.16666666))

        self._test_cases(cases)

    def test_multilabel_balanced(self):
        cases = []
        sol = np.array([[0, 0], [0, 0], [1, 1], [1, 1]])
        pred = np.array([[0, 0], [0, 0], [1, 1], [1, 1]])

        cases.append(('perfect', sol, pred, 1.0))
        cases.append(('completely wrong', sol, 1 - pred, 0.0))

        pred = np.array([[0, 0], [0, 0], [0, 0], [1, 1]])
        cases.append(('one sample per label wrong', sol, pred, 0.75))

        pred = np.array([[0, 0], [0, 0], [0, 1], [1, 1]])
        cases.append(('one sample in one label wrong', sol, pred, 0.875))

        pred = np.array([[0, 0], [0, 0], [0, 1], [0, 1]])
        cases.append(('two samples in one label wrong', sol, pred, 0.75))
        self._test_cases(cases)

    def test_cases_multilabel(self):
        cases = []
        num = 2

        sol = np.array([[1, 1, 1], [0, 0, 0], [0, 0, 0], [0, 0, 0]])
        sol3 = sol[:, 0:num]
        if num == 1:
            sol3 = np.array([sol3[:, 0]]).transpose()

        cases.append(('{} labels perfect'.format(num), sol3, sol3, 1.0))

        cases.append(('All wrong, in the multi-label sense', sol3, 1 - sol3,
                      0.0))

        sol4 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 1]])
        cases.append(('Three labels perfect', sol4, sol4, 1.0))

        cases.append(('Three classes all wrong, in the multi-label sense',
                      sol4, 1 - sol4, 0.0))

        self._test_cases(cases)


class PACTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def _test_cases(self, cases):
        for case in cases:
            testname, sol, pred, result = case

            pred = pred.astype(np.float32)
            sol, pred = copy_and_preprocess_arrays(sol, pred)
            pac = pac_score(sol, pred)
            self.assertAlmostEqual(pac, result, msg=testname, places=1)

    def test_cases_binary_score_verification(self):
        cases = []
        sol = np.array([0, 0, 1, 1])
        pred = np.array([[1, 0], [1, 0], [0, 1], [0, 1]])

        cases.append(('perfect', sol, pred, 1.0))
        cases.append(('anti-perfect', sol, 1 - pred, -1.0,))

        uneven_proba = np.array(
            [[0.7, 0.3], [0.4, 0.6], [0.49, 0.51], [0.2, 0.8]])

        cases.append(('uneven proba', sol, uneven_proba, 0.162745170342))

        eps = 1.e-15
        ties = np.array([[0.5 + eps, 0.5 - eps], [0.5 - eps, 0.5 + eps],
                         [0.5 + eps, 0.5 - eps], [0.5 - eps, 0.5 + eps]])
        cases.append(('ties_broken', sol, ties, 0.0))

        ties = np.array([[0.5, 0.5], [0.5, 0.5], [0.5, 0.5], [0.5, 0.5]])
        cases.append(('ties', sol, ties, 0.0))

        sol = np.array([0, 1, 1])
        pred = np.array([[0.5, 0.5], [0.5, 0.5], [0.5, 0.5]])
        cases.append(('even proba', sol, pred, -0.0618725166757))

        _pred = np.array([[1.0, 0.0], [0.0, 1.0], [0.0, 1.0]])
        pred = np.array([sum(_pred) * 1. / len(_pred)] * len(_pred))
        cases.append(('correct PAC prior', sol, pred, 0.0))

        pred = np.array([[1., 1.], [1., 1.], [1., 1.]])
        cases.append(('all positive', sol, pred, -1.12374503314))

        pred = np.array([[0, 0], [0, 0], [0, 0]])
        cases.append(('all negative', sol, pred, -1.1237237959))

        self._test_cases(cases)

    def test_cases_multiclass_score_verification(self):
        cases = []
        sol = np.array([0, 1, 2, 0])
        pred = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 2], [1, 0, 0]])

        cases.append(('3 classes perfect', sol, pred, 1.0))

        pred = np.array([[0, 1, 0], [0, 0, 1], [0, 1, 0], [0, 0, 1]])
        cases.append(('all classes wrong', sol, pred, -0.5469181142705154))

        pred = np.array([[0., 0., 0.]] * 4)
        cases.append(('equi proba (wrong test from the starting kit)',
                      sol, pred, -0.5469181142705154))

        pred = np.array([[1. / 3, 1. / 3, 1. / 3], [1. / 3, 1. / 3, 1. / 3],
                         [1. / 3, 1. / 3, 1. / 3], [1. / 3, 1. / 3, 1. / 3]])
        cases.append(('equi proba', sol, pred, -0.031278784012588157))

        pred = np.array([[0.2, 0, 0.5], [0.8, 0.4, 0.1], [0.9, 0.1, 0.2],
                         [0.7, 0.3, 0.3]])
        cases.append(('sum(proba) < 1.0', sol, pred, -0.085886926180064257))

        pred = np.array([[0.5, 0.25, 0.25], [0.5, 0.25, 0.25],
                         [0.5, 0.25, 0.25], [0.5, 0.25, 0.25]])
        cases.append(('predict prior', sol, pred, 0))

        self._test_cases(cases)

    def test_cases_multilabel_1l(self):
        cases = []
        num = 2

        sol = np.array([[1, 1, 1], [0, 0, 0], [0, 0, 0], [0, 0, 0]])
        sol3 = sol[:, 0:num]
        if num == 1:
            sol3 = np.array([sol3[:, 0]]).transpose()

        cases.append(('{} labels perfect'.format(num), sol3, sol3, 1.0))

        cases.append(('All wrong, in the multi-label sense', sol3, 1 - sol3,
                      -1.32491508679))

        pred = np.array([[0.5, 0.5, 0.5], [0.5, 0.5, 0.5], [0.5, 0.5, 0.5],
                         [0.5, 0.5, 0.5]])
        if num == 1:
            pred = np.array([pred[:, 0]]).transpose()
        else:
            pred = pred[:, 0:num]
        cases.append(('All equi proba: 0.5', sol3, pred, -0.162457543395))

        pred = np.array(
            [[0.25, 0.25, 0.25], [0.25, 0.25, 0.25], [0.25, 0.25, 0.25],
             [0.25, 0.25, 0.25]])
        if num == 1:
            pred = np.array([pred[:, 0]]).transpose()
        else:
            pred = pred[:, 0:num]
        cases.append(('All equi proba, prior: 0.25', sol3, pred, 0.0))

        pred = np.array([[0.2, 0.2, 0.2], [0.8, 0.8, 0.8], [0.9, 0.9, 0.9],
                         [0.7, 0.7, 0.7]])
        if num == 1:
            pred = np.array([pred[:, 0]]).transpose()
        else:
            pred = pred[:, 0:num]
        cases.append(('Some proba', sol3, pred, -0.892199631436))

        pred = np.array([[0.2, 0.2, 0.2], [0.8, 0.8, 0.8], [0.9, 0.9, 0.9],
                         [0.7, 0.7, 0.7]])
        if num == 1:
            pred = np.array([pred[:, 0]]).transpose()
        else:
            pred = pred[:, 0:num]
        cases.append(('Invert both solution and prediction', 1 - sol3, pred,
                      0.5277086603))

        self._test_cases(cases)

    def test_cases_multilabel_2(self):
        cases = []

        sol4 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 1]])
        cases.append(('Three labels perfect', sol4, sol4, 1.0))

        cases.append(('Three classes all wrong, in the multi-label sense',
                      sol4, 1 - sol4, -1.20548265539))

        # Not at random because different classes have different priors
        pred = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]])
        cases.append(('Three classes equi proba (wrong test from StartingKit)',
                      sol4, pred, -1.20522116785))

        pred = np.array([[1. / 3, 1. / 3, 1. / 3], [1. / 3, 1. / 3, 1. / 3],
                         [1. / 3, 1. / 3, 1. / 3], [1. / 3, 1. / 3, 1. / 3]])
        cases.append(('Three classes equi proba', sol4, pred,
                      -0.034665665346400684))

        pred = np.array([[0.2, 0, 0.5], [0.8, 0.4, 0.1], [0.9, 0.1, 0.2],
                         [0.7, 0.3, 0.3]])
        cases.append(('Three classes some proba that do not add up', sol4,
                      pred, -0.249775129382))

        pred = np.array([[0.25, 0.25, 0.5], [0.25, 0.25, 0.5],
                         [0.25, 0.25, 0.5], [0.25, 0.25, 0.5]])
        cases.append(('Three classes predict prior', sol4, pred, 0.0))

        self._test_cases(cases)


from collections import defaultdict
import itertools
import logging
import os
import Queue
import time

import numpy as np
import pandas as pd
import sklearn
import scipy.stats

from autosklearn.metalearning.metalearning.meta_base import MetaBase
import HPOlib.benchmark_util as benchmark_util
from autosklearn.metalearning.metalearning.kNearestDatasets.kND import \
    LearnedDistanceRF

logger = logging.getLogger(__name__)


# ###############################################################################
# Stuff for offline hyperparameter optimization of the Distance RF

def _validate_rf_without_one_dataset(X, Y, rf, task_id):
    X_train, Y_train, X_valid, Y_valid = split_for_loo(X, Y, task_id)
    predictions = rf.model.predict(X_valid)
    rho = scipy.stats.kendalltau(Y_valid, predictions)[0]
    mae = sklearn.metrics.mean_absolute_error(predictions, Y_valid)
    mse = sklearn.metrics.mean_squared_error(predictions, Y_valid)
    return mae, mse, rho


def train_rf_without_one_dataset(X, Y, rf, task_id):
    # Pay attention, this is not for general sklearn models, but for adapted
    # models...
    X_train, Y_train, X_valid, Y_valid = split_for_loo(X, Y, task_id)
    rf.model.fit(X_train, Y_train)
    return rf


def split_for_loo(X, Y, task_id):
    train = []
    valid = []
    for cross in X.index:
        if str(task_id) in cross:
            valid.append(cross)
        else:
            train.append(cross)

    X_train = X.loc[train].values
    Y_train = Y.loc[train].values.reshape((-1,))
    X_valid = X.loc[valid].values
    Y_valid = Y.loc[valid].values.reshape((-1,))
    return X_train, Y_train, X_valid, Y_valid


# TODO: this file has too many tasks, move the optimization of the metric
# function and the forward selection to some different files,
# maybe generalize these things to work for other models as well...
if __name__ == "__main__":
    """For a given problem train the metric function and return its loss
    value. Arguments:
      * task_files_list
      * experiment_files_list
      * metalearning_directory

    You can also enable forward selection by adding '--forward_selection True'
    You can also enable embedded feature selection by adding '--embedded_selection True'
    You can add '--keep_configurations -preprocessing=None,-classifier=LibSVM
    Sample call: python kND.py --task_files_list /home/feurerm/thesis/experiments/experiment/2014_06_01_AutoSklearn_metalearning/tasks.txt
    --experiments_list /home/feurerm/thesis/experiments/experiment/2014_06_01_AutoSklearn_metalearning/experiments_fold0.txt
    --metalearning_directory /home/feurerm/thesis/experiments/experiment --params -random_state 5
    """
    starttime = time.time()
    args, params = benchmark_util.parse_cli()
    os.chdir(args['metalearning_directory'])
    #pyMetaLearn.directory_manager.set_local_directory(
    #    args['metalearning_directory'])

    with open(args["task_files_list"]) as fh:
        task_files_list = fh.readlines()
    with open(args["experiments_list"]) as fh:
        experiments_list = fh.readlines()

    if 'keep_configurations' in args:
        keep_configurations = args['keep_configurations']
        keep_configurations = keep_configurations.split(',')
        keep_configurations = tuple(
            [tuple(kc.split('=')) for kc in keep_configurations])
    else:
        keep_configurations = None

    meta_base = MetaBase(task_files_list, experiments_list, keep_configurations)
    metafeatures = meta_base.get_all_train_metafeatures_as_pandas()
    runs = meta_base.get_all_runs()

    # This can print the best hyperparameters of every dataset
    # for dataset in runs:
    # print dataset, sorted(runs[dataset], key=lambda t: t.result)[0]

    rf = LearnedDistanceRF(**params)
    X, Y = rf._create_dataset(metafeatures, runs)
    import cPickle

    with open("test.pkl", "w") as fh:
        cPickle.dump((X, Y, metafeatures), fh, -1)

    print("Metafeatures", metafeatures.shape)
    print("X", X.shape, np.isfinite(X).all().all())
    print("Y", Y.shape, np.isfinite(Y).all())

    metafeature_sets = Queue.Queue()
    if 'forward_selection' in args:
        used_metafeatures = []
        metafeature_performance = []
        print("Starting forward selection ",)
        i = 0
        for m1, m2 in itertools.combinations(metafeatures.columns, 2):
            metafeature_sets.put(pd.Index([m1, m2]))
            i += 1
        print("with %d metafeature combinations" % i)
    elif 'embedded_selection' in args:
        metafeature_performance = []
        metafeature_sets.put(metafeatures.columns)
    else:
        metafeature_sets.put(metafeatures.columns)

    while not metafeature_sets.empty():
        metafeature_set = metafeature_sets.get()
        metafeature_ranks = defaultdict(float)
        loo_mae = []
        loo_rho = []
        loo_mse = []

        print("###############################################################")
        print("New iteration of FS with:")
        print(metafeature_set)
        print("Dataset Mae MSE Rho")
        # Leave one out CV
        for idx in range(metafeatures.shape[0]):
            leave_out_dataset = metafeatures.index[idx]
            if 'forward_selection' not in args:
                print(leave_out_dataset,)

            columns = np.hstack(("0_" + metafeature_set,
                                 "1_" + metafeature_set))
            X_ = X.loc[:, columns]
            rf = train_rf_without_one_dataset(X_, Y, rf, leave_out_dataset)
            mae, mse, rho = _validate_rf_without_one_dataset(X_, Y, rf,
                                                             leave_out_dataset)
            if 'forward_selection' not in args:
                print(mae, mse, rho)

            loo_mae.append(mae)
            loo_rho.append(rho)
            loo_mse.append(mse)
            mf_importances = [(rf.model.feature_importances_[i], X_.columns[i])
                              for i in range(X_.shape[1])]
            mf_importances.sort()
            mf_importances.reverse()
            for rank, item in enumerate(mf_importances):
                score, mf_name = item
                metafeature_ranks[mf_name] += float(rank)

        mae = np.mean(loo_mae)
        mae_std = np.std(loo_mae)
        mse = np.mean(loo_mse)
        mse_std = np.mean(loo_mse)
        rho = np.mean(loo_rho)
        rho_std = np.std(loo_rho)

        mean_ranks = [
            (metafeature_ranks[mf_name] / metafeatures.shape[0], mf_name)
            for mf_name in X.columns]
        mean_ranks.sort()

        # TODO: save only validate-best runs!
        print("MAE", mae, mae_std)
        print("MSE", mse, mse_std)
        print("Mean tau", rho, rho_std)
        duration = time.time() - starttime

        if 'forward_selection' in args:
            metafeature_performance.append((mse, metafeature_set))
            # TODO: this can also be sorted in a pareto-optimal way...

            if metafeature_sets.empty():
                if len(used_metafeatures) == 10:
                    break

                print("#######################################################")
                print("#######################################################")
                print("Adding a new feature to the feature set")
                metafeature_performance.sort()
                print(metafeature_performance)
                used_metafeatures = metafeature_performance[0][1]
                for metafeature in metafeatures.columns:
                    if metafeature in used_metafeatures:
                        continue
                    # I don't know if indexes are copied
                    tmp = [uaie for uaie in used_metafeatures]
                    tmp.append(metafeature)
                    metafeature_sets.put(pd.Index(tmp))
                metafeature_performance = []

        elif 'embedded_selection' in args:
            if len(metafeature_set) <= 2:
                break

            # Remove a metafeature; elements are (average rank, name);
            # only take the name from index two on
            # because the metafeature is preceeded by the index of the
            # dataset which is either 0_ or 1_
            remove = mean_ranks[-1][1][2:]
            print("Going to remove", remove)
            keep = pd.Index([mf_name for mf_name in metafeature_set if
                             mf_name != remove])
            print("I will keep", keep)
            metafeature_sets.put(keep)

        else:
            for rank in mean_ranks:
                print(rank)

    if 'forward_selection' in args:
        metafeature_performance.sort()
        print(metafeature_performance)
        mse = metafeature_performance[0][0]
    print("Result for ParamILS: %s, %f, 1, %f, %d, %s" % \
          ("SAT", abs(duration), mse, -1, str(__file__)))


import numpy as np
from scipy import sparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array


def _transform_selected(X, transform, selected="all", copy=True):
    """Apply a transform function to portion of selected features

    Parameters
    ----------
    X : array-like or sparse matrix, shape=(n_samples, n_features)
        Dense array or sparse matrix.

    transform : callable
        A callable transform(X) -> X_transformed

    copy : boolean, optional
        Copy X even if it could be avoided.

    selected: "all" or array of indices or mask
        Specify which features to apply the transform to.

    Returns
    -------
    X : array or sparse matrix, shape=(n_samples, n_features_new)
    """
    X = check_array(X, accept_sparse='csc', force_all_finite=False, copy=copy)

    if selected == "all":
        return transform(X)

    if len(selected) == 0:
        return X

    n_features = X.shape[1]
    ind = np.arange(n_features)
    sel = np.zeros(n_features, dtype=bool)
    sel[np.asarray(selected)] = True
    not_sel = np.logical_not(sel)
    n_selected = np.sum(sel)

    if n_selected == 0:
        # No features selected.
        return X
    elif n_selected == n_features:
        # All features selected.
        return transform(X)
    else:
        X_sel = transform(X[:, ind[sel]])
        X_not_sel = X[:, ind[not_sel]]

        if sparse.issparse(X_sel) or sparse.issparse(X_not_sel):
            # This is pretty memory-intense, making the memory usage for OpenML
            # task 146810 go over 3GB
            return sparse.hstack((X_sel, X_not_sel), format='csr')
        else:
            return np.hstack((X_sel, X_not_sel))


class OneHotEncoder(BaseEstimator, TransformerMixin):
    """Encode categorical integer features using a one-hot aka one-of-K scheme.

    The input to this transformer should be a matrix of integers, denoting
    the values taken on by categorical (discrete) features. The output will be
    a sparse matrix were each column corresponds to one possible value of one
    feature. It is assumed that input features take on values in the range
    [0, n_values).

    This encoding is needed for feeding categorical data to many scikit-learn
    estimators, notably linear models and SVMs with the standard kernels.

    Parameters
    ----------

    categorical_features: "all" or array of indices or mask
        Specify what features are treated as categorical.

        - 'all' (default): All features are treated as categorical.
        - array of indices: Array of categorical feature indices.
        - mask: Array of length n_features and with dtype=bool.

        Non-categorical features are always stacked to the right of the matrix.

    dtype : number type, default=np.float
        Desired dtype of output.

    sparse : boolean, default=True
        Will return sparse matrix if set True else will return an array.

    Attributes
    ----------
    `active_features_` : array
        Indices for active features, meaning values that actually occur
        in the training set. Only available when n_values is ``'auto'``.

    `feature_indices_` : array of shape (n_features,)
        Indices to feature ranges.
        Feature ``i`` in the original data is mapped to features
        from ``feature_indices_[i]`` to ``feature_indices_[i+1]``
        (and then potentially masked by `active_features_` afterwards)

    `n_values_` : array of shape (n_features,)
        Maximum number of values per feature.

    Examples
    --------
    Given a dataset with three features and two samples, we let the encoder
    find the maximum value per feature and transform the data to a binary
    one-hot encoding.

    >>> from sklearn.preprocessing import OneHotEncoder
    >>> enc = OneHotEncoder()
    >>> enc.fit([[0, 0, 3], [1, 1, 0], [0, 2, 1], \
[1, 0, 2]])  # doctest: +ELLIPSIS
    OneHotEncoder(categorical_features='all', dtype=<... 'float'>,
           sparse=True, minimum_fraction=None)
    >>> enc.n_values_
    array([2, 3, 4])
    >>> enc.feature_indices_
    array([0, 2, 5, 9])
    >>> enc.transform([[0, 1, 1]]).toarray()
    array([[ 1.,  0.,  0.,  1.,  0.,  0.,  1.,  0.,  0.]])

    See also
    --------
    sklearn.feature_extraction.DictVectorizer : performs a one-hot encoding of
      dictionary items (also handles string-valued features).
    sklearn.feature_extraction.FeatureHasher : performs an approximate one-hot
      encoding of dictionary items or strings.
    """

    def __init__(self, categorical_features="all", dtype=np.float,
                 sparse=True, minimum_fraction=None):
        self.categorical_features = categorical_features
        self.dtype = dtype
        self.sparse = sparse
        self.minimum_fraction = minimum_fraction

    def fit(self, X, y=None):
        """Fit OneHotEncoder to X.

        Parameters
        ----------
        X : array-like, shape=(n_samples, n_feature)
            Input array of type int.

        Returns
        -------
        self
        """
        self.fit_transform(X)
        return self

    def _fit_transform(self, X):
        """Assumes X contains only categorical features."""

        # First increment everything by three to account for the fact that
        # np.NaN will get an index of two, and 'other' values will get index of
        # one, index of zero is not assigned to also work with sparse data
        if sparse.issparse(X):
            X.data += 3
            X.data[~np.isfinite(X.data)] = 2
        else:
            X += 3
            X[~np.isfinite(X)] = 2

        X = check_array(X, accept_sparse='csc', force_all_finite=False,
                        dtype=np.int32)

        if X.min() < 0:
            raise ValueError("X needs to contain only non-negative integers.")
        n_samples, n_features = X.shape

        # Remember which values should not be replaced by the value 'other'
        if self.minimum_fraction is not None:
            do_not_replace_by_other = list()
            for column in range(X.shape[1]):
                do_not_replace_by_other.append(set())

                if sparse.issparse(X):
                    indptr_start = X.indptr[column]
                    indptr_end = X.indptr[column + 1]
                    unique, counts = np.unique(
                        X.data[indptr_start:indptr_end], return_counts=True,
                    )
                    colsize = indptr_end - indptr_start
                else:
                    unique, counts = np.unique(
                        X[:, column], return_counts=True,
                    )
                    colsize = X.shape[0]

                for unique_value, count in zip(unique, counts):

                    fraction = float(count) / colsize
                    if fraction >= self.minimum_fraction:
                        do_not_replace_by_other[-1].add(unique_value)

                for unique_value in unique:
                    if unique_value not in do_not_replace_by_other[-1]:
                        if sparse.issparse(X):
                            indptr_start = X.indptr[column]
                            indptr_end = X.indptr[column + 1]
                            X.data[indptr_start:indptr_end][
                                X.data[indptr_start:indptr_end] ==
                                unique_value] = 1
                        else:
                            X[:, column][X[:, column] == unique_value] = 1

            self.do_not_replace_by_other_ = do_not_replace_by_other

        if sparse.issparse(X):
            n_values = X.max(axis=0).toarray().flatten() + 2
        else:
            n_values = np.max(X, axis=0) + 2

        self.n_values_ = n_values
        n_values = np.hstack([[0], n_values])
        indices = np.cumsum(n_values)
        self.feature_indices_ = indices

        if sparse.issparse(X):
            row_indices = X.indices
            column_indices = []
            for i in range(len(X.indptr) - 1):
                nbr = X.indptr[i+1] - X.indptr[i]
                column_indices_ = [indices[i]] * nbr
                column_indices_ += X.data[X.indptr[i]:X.indptr[i+1]]
                column_indices.extend(column_indices_)
            data = np.ones(X.data.size)
        else:
            column_indices = (X + indices[:-1]).ravel().astype(np.int32)
            row_indices = np.repeat(np.arange(n_samples, dtype=np.int32),
                                    n_features)
            data = np.ones(n_samples * n_features, dtype=np.int32)

        out = sparse.coo_matrix((data, (row_indices, column_indices)),
                                shape=(n_samples, indices[-1]),
                                dtype=np.int32).tocsc()

        mask = np.array(out.sum(axis=0)).ravel() != 0
        active_features = np.where(mask)[0]
        out = out[:, active_features]
        self.active_features_ = active_features
        return out.tocsr() if self.sparse else out.toarray()

    def fit_transform(self, X, y=None):
        """Fit OneHotEncoder to X, then transform X.

        Equivalent to self.fit(X).transform(X), but more convenient and more
        efficient. See fit for the parameters, transform for the return value.
        """
        return _transform_selected(X, self._fit_transform,
                                   self.categorical_features, copy=True)

    def _transform(self, X):
        """Asssumes X contains only categorical features."""

        # First increment everything by three to account for the fact that
        # np.NaN will get an index of two, and 'other' values will get index of
        #  one, index of zero is not assigned to also work with sparse data
        if sparse.issparse(X):
            X.data += 3
            X.data[~np.isfinite(X.data)] = 2
        else:
            X += 3
            X[~np.isfinite(X)] = 2

        X = check_array(X, accept_sparse='csc', force_all_finite=False,
                        dtype=np.int32)
        if X.min() < 0:
            raise ValueError("X needs to contain only non-negative integers.")
        n_samples, n_features = X.shape

        indices = self.feature_indices_
        if n_features != indices.shape[0] - 1:
            raise ValueError("X has different shape than during fitting."
                             " Expected %d, got %d."
                             % (indices.shape[0] - 1, n_features))

        # Replace all indicators which were below `minimum_fraction` in the 
        # training set by 'other'
        if self.minimum_fraction is not None:
            for column in range(X.shape[1]):
                if sparse.issparse(X):
                    indptr_start = X.indptr[column]
                    indptr_end = X.indptr[column + 1]
                    unique = np.unique(X.data[indptr_start:indptr_end])
                else:
                    unique = np.unique(X[:, column])

                for unique_value in unique:
                    if unique_value not in self.do_not_replace_by_other_[
                        column]:
                        if sparse.issparse(X):
                            indptr_start = X.indptr[column]
                            indptr_end = X.indptr[column + 1]
                            X.data[indptr_start:indptr_end][
                                X.data[indptr_start:indptr_end] ==
                                unique_value] = 1
                        else:
                            X[:, column][X[:, column] == unique_value] = 1

        if sparse.issparse(X):
            n_values_check = X.max(axis=0).toarray().flatten() + 1
        else:
            n_values_check = np.max(X, axis=0) + 1

        # Replace all indicators which are out of bounds by 'other' (index 0)
        if (n_values_check > self.n_values_).any():
            # raise ValueError("Feature out of bounds. Try setting n_values.")
            for i, n_value_check in enumerate(n_values_check):
                if (n_value_check - 1) >= self.n_values_[i]:
                    if sparse.issparse(X):
                        indptr_start = X.indptr[i]
                        indptr_end = X.indptr[i+1]
                        X.data[indptr_start:indptr_end][X.data
                            [indptr_start:indptr_end] >= self.n_values_[i]] = 0
                    else:
                        X[:, i][X[:, i] >= self.n_values_[i]] = 0

        if sparse.issparse(X):
            row_indices = X.indices
            column_indices = []
            for i in range(len(X.indptr) - 1):
                nbr = X.indptr[i + 1] - X.indptr[i]
                column_indices_ = [indices[i]] * nbr
                column_indices_ += X.data[X.indptr[i]:X.indptr[i + 1]]
                column_indices.extend(column_indices_)
            data = np.ones(X.data.size)
        else:
            column_indices = (X + indices[:-1]).ravel().astype(np.int32)
            row_indices = np.repeat(np.arange(n_samples, dtype=np.int32),
                                    n_features)
            data = np.ones(n_samples * n_features, dtype=np.int32)
        out = sparse.coo_matrix((data, (row_indices, column_indices)),
                                shape=(n_samples, indices[-1]),
                                dtype=np.int32).tocsc()

        out = out[:, self.active_features_]
        return out.tocsr() if self.sparse else out.toarray()

    def transform(self, X):
        """Transform X using one-hot encoding.

        Parameters
        ----------
        X : array-like, shape=(n_samples, n_features)
            Input array of type int.

        Returns
        -------
        X_out : sparse matrix if sparse=True else a 2-d array, dtype=int
            Transformed input.
        """
        return _transform_selected(X, self._transform,
                                   self.categorical_features, copy=True)



# -*- encoding: utf-8 -*-
cimport numpy as np
import numpy as np

import scipy.sparse


from libc.stdio cimport *


def log_function(*args):
    print(args)

'''    read_sparse_file (filename, num_features, num_points)

    Reads the data from a file containing sparse data into a scipy.sparse.csr_matrix (type float32).
    
    The number of features and the number of points have to be known ahead of time.
    Each pair is "index:value" and the row in the file equals the row in the matrix.
    The labels are assumed to be in a different file, in contrast to the SVMlight file format where the "target value" would be the first entry    in each line!
    The optional argument initial_length should estimate the number of elements    expected in the data set. If it is to small all arrays will be dynamically enlarged in steps of initial_length. If chosen to small, unnecessary reallocations slow things down. If to big, memory might be wasted while    loading the data.
    The optional argument offset should be -1 if the indices in the file start at 1! This is the case for the automl challenge, hence the default value.
    So far, only ascii #32 (space) is recognized as a whitespace. If the entries are tab-separated (or any other chararcter), this could easily be implemented here.
    
'''
def read_sparse_file(filename, int num_points,int num_features, int initial_length = 8192, int offset = -1, double max_memory_in_mb = 1048576):

    #cdef np.ndarray[float, ndim=1] 
    data = np.zeros(initial_length,dtype=np.float32)
    #cdef np.ndarray[int, ndim=1]
    indices = np.zeros(initial_length, dtype=np.int32)
    #cdef np.ndarray[int, ndim=1] 
    indptr = np.zeros(num_points+1, dtype=np.int32)

    # we have to dynamically enlarge the arrays, so we need to keep track of how many entries we already have
    cdef int num_entries = 0
    
    
    # variables for I/O
    cdef char* fname
    cdef FILE* cfile
    cdef ssize_t read
    cdef char whitespace

    #variables for the indices and values
    cdef int i=0,j
    cdef float v


    filename_byte_string = filename.encode("UTF-8")
    fname = filename_byte_string
    cfile = fopen(fname, "r")
    if cfile == NULL:
        raise RuntimeError("Couldn't find file {}".format(filename))

    while True:
        # read the column and the value and store it
        read =fscanf(cfile, "%i:%f",&j,&v)
        # stop at EOF
        if read == -1:
            break

        data[num_entries] = v
        indices[num_entries] = j+offset
        num_entries += 1
        #enlarge the array if necessary
        if num_entries == data.shape[0]:
            if ((data.nbytes + indices.nbytes) < max_memory_in_mb*1024*1024):
                data.resize(data.shape[0]+initial_length)
                indices.resize(data.shape[0])
            else:
                break
        
        # check if we hit a endline next to recognize the next row
        # It is cumbersome, but a way to reliably do it!
        whitespace = fgetc(cfile)
        while whitespace == 32:
            whitespace = fgetc(cfile)

        if whitespace == '\n':
            i+=1
            # stop if num_points have been read
            if i >= num_points:
                break
            indptr[i] = num_entries
        else:
            ungetc(whitespace, cfile)

    fclose(cfile)
    
    
    #cut arrays to size
    data.resize(num_entries)
    indices.resize(num_entries)
    
    # fix the end of indptr
    for j in range (i,num_points+1):
        indptr[j] = num_entries
    return (scipy.sparse.csr_matrix((data, indices, indptr),
                                    shape=[num_points, num_features]))



'''
    see read_sparse_file, only difference: the value of every index present is 1, so there are no index:value pairs, but just indices.
    
'''
def read_sparse_binary_file(filename, int num_points, int num_features, int initial_length = 8192, int offset = -1, double max_memory_in_mb = 1048576):

    data = np.zeros(initial_length,dtype=np.bool)
    indices = np.zeros(initial_length, dtype=np.int32)
    cdef np.ndarray[int, ndim=1] indptr = np.zeros(num_points+1, dtype=np.int32)
    
    # we have to dynamically enlarge the arrays, so we need to keep track of how many entries we already have
    cdef int num_entries = 0
    
    # variables for I/O
    cdef char* fname
    cdef FILE* cfile
    cdef ssize_t read
    cdef char whitespace

    #variables for the indices and values
    cdef int i=0,j

    filename_byte_string = filename.encode("UTF-8")
    fname = filename_byte_string
    cfile = fopen(fname, "r")
    if cfile == NULL:
        raise RuntimeError("Couldn't find file {}".format(filename))

    while True:
        # read the column and the value and store it
        read =fscanf(cfile, "%d",&j)
        # stop at EOF
        if read == -1:
            break

        data[num_entries] = True
        indices[num_entries] = j+offset
        num_entries += 1
        
        #enlarge the array if necessary
        if num_entries == data.shape[0]:
            if ((data.nbytes + indices.nbytes) < max_memory_in_mb*1024*1024):
                data.resize(data.shape[0]+initial_length)
                indices.resize(data.shape[0])
            else:
                break
        
        # check if we hit a endline next to recognize the next row
        # It is cumbersome, but a way to reliably do it!
        whitespace = fgetc(cfile)
        while (whitespace==32):
            whitespace = fgetc(cfile)

        if whitespace == '\n':
            i+=1
            # stop if num_points have been read
            if (i >=num_points): break
            indptr[i] = num_entries
        else:
            ungetc(whitespace, cfile)

    fclose(cfile)
    
    
    #cut arrays to size
    data.resize(num_entries)
    indices.resize(num_entries)
    
    # fix the end of indptr
    for j in range (i,num_points+1):
        indptr[j] = num_entries

    return (scipy.sparse.csr_matrix((data, indices, indptr),
                                    shape=[num_points, num_features],
                                    dtype=np.bool))


'''    read_dense_file (filename, num_features, num_points)

    Reads the data from a file containing dense data into a numpy array (type float32)
    
    The number of features and the number of points have to be known ahead of time.
    
    The function does not check for EOF or missing values, so be cautious!
'''
def read_dense_file(filename, int num_points, int num_features,double max_memory_in_mb = 1048576):

    nbits = np.finfo(np.float32).nexp + np.finfo(np.float32).nmant+1
    num_points = long(min(num_points,max_memory_in_mb*1024*1024*8/nbits/num_features))
    
    cdef np.ndarray[float, ndim=2] data = np.zeros([num_points, num_features],dtype=np.float32)
    
    # variables for I/O
    cdef char* fname
    cdef FILE* cfile

    #variable for the indices and values
    cdef int i=0,j=0
    cdef float v

    filename_byte_string = filename.encode("UTF-8")
    fname = filename_byte_string
    cfile = fopen(fname, "r")
    if cfile == NULL:
        raise RuntimeError("Couldn't find file {}".format(filename))

    for i in range(num_points):
        for j in range(num_features):
            fscanf(cfile, "%f",&v)
            data[i,j] = v
    fclose(cfile)

    return(data)


def read_dense_file_unknown_width(filename, num_points, max_memory_in_mb = 1048576):
    # variables for I/O
    cdef char* fname
    cdef FILE* cfile
    cdef int rc;

    #variables for the indices and values
    cdef int num_cols=0

    filename_byte_string = filename.encode("UTF-8")
    fname = filename_byte_string
    cfile = fopen(fname, "r")
    if cfile == NULL:
        raise RuntimeError("Couldn't find file {}".format(filename))


    #count the number of columns in the first line
    rc = fgetc(cfile)
    while True:
        # read a column
        while (rc != ' ') and (rc != '\n'):
            rc=fgetc(cfile)
        
        while (rc == ' '):
            rc=fgetc(cfile)
        num_cols+=1
        
        if (rc == '\n'): break
    fclose(cfile)

    data = read_dense_file(filename, num_points,num_cols, max_memory_in_mb)

    # if only one predictor is present, convert it into a 1D array
    if data.shape[1] == 1:
        return (data.flatten())

    return (data)


# function copied from the reference implementation
# no need to really optimize them because they don't take long

def read_first_line (filename):
    """
    Read fist line of file
    :param filename:
    :return:
    """
    data =[]
    with open(filename, "r") as data_file:
        line = data_file.readline()
        data = line.strip().split()
    return data  


def file_to_array (filename, verbose=False):
    """
    Converts a file to a list of list of STRING
    It differs from np.genfromtxt in that the number of columns doesn't need to be constant
    :param filename:
    :param verbose:
    :return:
    """

    data =[]
    with open(filename, "r") as data_file:
        if verbose:
            log_function("Reading {}...".format(filename))
        lines = data_file.readlines()

        if verbose:
            log_function("Converting {} to correct array...".format(filename))

        data = [lines[i].strip().split() for i in range (len(lines))]
    return data



import numpy as np

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import UniformFloatHyperparameter, \
    UniformIntegerHyperparameter, UnParametrizedHyperparameter,  \
    CategoricalHyperparameter, Constant
from ConfigSpace.conditions import EqualsCondition

from autosklearn.pipeline.components.base import (
    AutoSklearnClassificationAlgorithm,
    IterativeComponentWithSampleWeight,
)
from autosklearn.pipeline.implementations.xgb import CustomXGBClassifier
from autosklearn.pipeline.constants import *


class XGradientBoostingClassifier(
    IterativeComponentWithSampleWeight,
    AutoSklearnClassificationAlgorithm,
):
    def __init__(self,
        # General Hyperparameters
        learning_rate, n_estimators, subsample, booster, max_depth,
        colsample_bylevel, colsample_bytree, reg_alpha, reg_lambda,
        # Inactive Hyperparameters
        gamma, min_child_weight, max_delta_step,
        base_score, scale_pos_weight, n_jobs=1, init=None,
        random_state=None, verbose=0,
        # (Conditional) DART Hyperparameters
        sample_type=None, normalize_type=None, rate_drop=None,
    ):

        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.subsample = subsample
        self.max_depth = max_depth
        self.booster = booster

        booster_args = {
            'sample_type': sample_type,
            'normalize_type': normalize_type,
            'rate_drop': rate_drop,
        }
        if any(v is not None for v in booster_args.values()):
            self.booster_args = booster_args
        else:
            self.booster_args = {}

        ## called differently
        # max_features: Subsample ratio of columns for each split, in each level.
        self.colsample_bylevel = colsample_bylevel

        # min_weight_fraction_leaf: Minimum sum of instance weight(hessian)
        # needed in a child.
        self.min_child_weight = min_child_weight

        # Whether to print messages while running boosting.
        if verbose:
            self.silent = False
        else:
            self.silent = True

        # Random number seed.
        if random_state is None:
            self.seed = 1
        else:
            self.seed = random_state.randint(1, 10000, size=1)[0]

        ## new paramaters
        # Subsample ratio of columns when constructing each tree.
        self.colsample_bytree = colsample_bytree

        # Minimum loss reduction required to make a further partition on a leaf
        # node of the tree.
        self.gamma = gamma

        # Maximum delta step we allow each tree's weight estimation to be.
        self.max_delta_step = max_delta_step

        # L2 regularization term on weights
        self.reg_alpha = reg_alpha

        # L1 regularization term on weights
        self.reg_lambda = reg_lambda

        # Balancing of positive and negative weights.
        self.scale_pos_weight = scale_pos_weight

        # The initial prediction score of all instances, global bias.
        self.base_score = base_score

        # Number of parallel threads used to run xgboost.
        self.n_jobs = n_jobs

        ## Were there before, didn't touch
        self.init = init
        self.estimator = None

    def iterative_fit(self, X, y, n_iter=2, refit=False, sample_weight=None):

        import sklearn.model_selection
        X_train, X_test, y_train, y_test = \
            sklearn.model_selection.train_test_split(
                X, y, random_state=1, test_size=0.1)

        if refit:
            self.estimator = None

        if self.estimator is None:
            self.learning_rate = float(self.learning_rate)
            self.n_estimators = int(self.n_estimators)
            self.subsample = float(self.subsample)
            self.max_depth = int(self.max_depth)

            # (TODO) Gb used at most half of the features, here we use all
            self.colsample_bylevel = float(self.colsample_bylevel)

            self.colsample_bytree = float(self.colsample_bytree)
            self.gamma = float(self.gamma)
            self.min_child_weight = int(self.min_child_weight)
            self.max_delta_step = int(self.max_delta_step)
            self.reg_alpha = float(self.reg_alpha)
            self.reg_lambda = float(self.reg_lambda)
            self.n_jobs = int(self.n_jobs)
            self.base_score = float(self.base_score)
            self.scale_pos_weight = float(self.scale_pos_weight)
            for key in self.booster_args:
                try:
                    self.booster_args[key] = float(self.booster_args[key])
                except:
                    pass

            # We don't support multilabel, so we only need 1 objective function
            if len(np.unique(y)) == 2:
                # We probably have binary classification
                # TODO: use logitraw ?
                self.objective = 'binary:logistic'
                self.eval_metric = 'error'
            else:
                self.objective = 'multi:softmax'
                self.eval_metric = 'merror'

            arguments = dict(
                max_depth=self.max_depth,
                learning_rate=self.learning_rate,
                n_estimators=n_iter,
                silent=self.silent,
                booster=self.booster,
                objective=self.objective,
                n_jobs=self.n_jobs,
                gamma=self.gamma,
                scale_pos_weight=self.scale_pos_weight,
                min_child_weight=self.min_child_weight,
                max_delta_step=self.max_delta_step,
                subsample=self.subsample,
                colsample_bytree=self.colsample_bytree,
                colsample_bylevel=self.colsample_bylevel,
                reg_alpha=self.reg_alpha,
                reg_lambda=self.reg_lambda,
                base_score=self.base_score,
                seed=self.seed,
                random_state=self.seed,
                **self.booster_args
            )

            self.estimator = CustomXGBClassifier(
                tree_method='auto',
                **arguments
            )
            self.estimator.fit(
                X_train, y_train,
                sample_weight=sample_weight,
                eval_set=[(X_test, y_test)],
                eval_metric=self.eval_metric,
                verbose=False,
            )

        elif not self.configuration_fully_fitted():
            n_estimators_before_fit = self.estimator.n_estimators
            self.estimator.n_estimators += n_iter
            self.estimator.n_estimators = min(self.estimator.n_estimators,
                                              self.n_estimators)
            self.estimator.fit(
                X_train, y_train,
                xgb_model=self.estimator.get_booster(),
                sample_weight=sample_weight,
                eval_set=[(X_train, y_train), (X_test, y_test)],
                early_stopping_rounds=int(self.n_estimators / 10),
                eval_metric=self.eval_metric,
                verbose=False,
            )
            if (
                n_estimators_before_fit + n_iter
                > self.estimator.best_iteration + int(self.n_estimators / 10)
            ) or (
                self.estimator.n_estimators >= self.n_estimators
            ):
                self.fully_fit_ = True

        return self

    def configuration_fully_fitted(self):
        if self.estimator is None:
            return False
        elif not hasattr(self, 'fully_fit_'):
            return False
        else:
            return self.fully_fit_

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    def predict_proba(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        return self.estimator.predict_proba(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {'shortname': 'XGB',
                'name': 'XGradient Boosting Classifier',
                'handles_regression': False,
                'handles_classification': True,
                'handles_multiclass': True,
                'handles_multilabel': False,
                'is_deterministic': True,
                'input': (DENSE, SPARSE, UNSIGNED_DATA),
                'output': (PREDICTIONS,)}

    @staticmethod
    def get_hyperparameter_search_space(dataset_properties=None):
        cs = ConfigurationSpace()

        # Parameterized Hyperparameters
        max_depth = UniformIntegerHyperparameter(
            name="max_depth", lower=1, upper=20, default_value=3
        )
        learning_rate = UniformFloatHyperparameter(
            name="learning_rate", lower=0.001, upper=1, default_value=0.1,
            log=True,
        )
        n_estimators = Constant("n_estimators", 512)
        booster = CategoricalHyperparameter(
            "booster", ["gbtree", "dart"]
        )
        subsample = UniformFloatHyperparameter(
            name="subsample", lower=0.01, upper=1.0, default_value=1.0, log=False
        )
        min_child_weight = UniformIntegerHyperparameter(
            name="min_child_weight", lower=0,
            upper=20, default_value=1, log=False
        )
        colsample_bytree = UniformFloatHyperparameter(
            name="colsample_bytree", lower=0.1, upper=1.0, default_value=1,
        )
        colsample_bylevel = UniformFloatHyperparameter(
            name="colsample_bylevel", lower=0.1, upper=1.0, default_value=1,
        )
        reg_alpha = UniformFloatHyperparameter(
            name="reg_alpha", lower=1e-10, upper=1e-1, log=True,
            default_value=1e-10)
        reg_lambda = UniformFloatHyperparameter(
            name="reg_lambda", lower=1e-10, upper=1e-1, log=True,
            default_value=1e-10)

        # DART Hyperparameters
        sample_type = CategoricalHyperparameter(
            'sample_type', ['uniform', 'weighted'], default_value='uniform',
        )
        normalize_type = CategoricalHyperparameter(
            'normalize_type', ['tree', 'forest'], default_value='tree',
        )
        rate_drop = UniformFloatHyperparameter(
            'rate_drop', 1e-10, 1-(1e-10), default_value=0.5,
        )

        # Unparameterized Hyperparameters
        # https://xgboost.readthedocs.io/en/latest//parameter.html
        # minimum loss reduction required to make a further partition on a
        # leaf node of the tree
        gamma = UnParametrizedHyperparameter(
            name="gamma", value=0)
        # absolute regularization (in contrast to eta), comparable to
        # gradient clipping in deep learning - according to the internet this
        #  is most important for unbalanced data
        max_delta_step = UnParametrizedHyperparameter(
            name="max_delta_step", value=0)
        base_score = UnParametrizedHyperparameter(
            name="base_score", value=0.5)
        scale_pos_weight = UnParametrizedHyperparameter(
            name="scale_pos_weight", value=1)

        cs.add_hyperparameters([
            # Active
            max_depth, learning_rate, n_estimators, booster,
            subsample, colsample_bytree, colsample_bylevel,
            reg_alpha, reg_lambda,
            # DART
            sample_type, normalize_type, rate_drop,
            # Inactive
            min_child_weight, max_delta_step, gamma,
            base_score, scale_pos_weight
        ])

        sample_type_condition = EqualsCondition(
            sample_type, booster, 'dart',
        )
        normalize_type_condition = EqualsCondition(
            normalize_type, booster, 'dart',
        )
        rate_drop_condition = EqualsCondition(
            rate_drop, booster, 'dart',
        )

        cs.add_conditions([
            sample_type_condition, normalize_type_condition,
            rate_drop_condition,
        ])
        return cs



import os
import sys

# Add the parent directory to the path to import the parent component as
# dummy_components.dummy_component_2.DummyComponent1
this_directory = os.path.dirname(os.path.abspath(__file__))
parent_directory = os.path.abspath(os.path.join(this_directory, '..'))
sys.path.append(parent_directory)

from autosklearn.pipeline.components.base import AutoSklearnClassificationAlgorithm
from dummy_components.dummy_component_1 import DummyComponent1


class DummyComponent2(DummyComponent1):
    pass




import copy
import resource
import sys
import traceback
import unittest
import unittest.mock

import numpy as np
import sklearn.datasets
import sklearn.decomposition
import sklearn.ensemble
import sklearn.svm
from sklearn.utils.testing import assert_array_almost_equal

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter

from autosklearn.pipeline.regression import SimpleRegressionPipeline
from autosklearn.pipeline.components.base import \
    AutoSklearnPreprocessingAlgorithm, AutoSklearnRegressionAlgorithm
import autosklearn.pipeline.components.regression as regression_components
import autosklearn.pipeline.components.feature_preprocessing as preprocessing_components
from autosklearn.pipeline.util import get_dataset
from autosklearn.pipeline.constants import *


class SimpleRegressionPipelineTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_io_dict(self):
        regressors = regression_components._regressors
        for r in regressors:
            if regressors[r] == regression_components.RegressorChoice:
                continue
            props = regressors[r].get_properties()
            self.assertIn('input', props)
            self.assertIn('output', props)
            inp = props['input']
            output = props['output']

            self.assertIsInstance(inp, tuple)
            self.assertIsInstance(output, tuple)
            for i in inp:
                self.assertIn(i, (SPARSE, DENSE, SIGNED_DATA, UNSIGNED_DATA))
            self.assertEqual(output, (PREDICTIONS,))
            self.assertIn('handles_regression', props)
            self.assertTrue(props['handles_regression'])
            self.assertIn('handles_classification', props)
            self.assertIn('handles_multiclass', props)
            self.assertIn('handles_multilabel', props)
            self.assertFalse(props['handles_classification'])
            self.assertFalse(props['handles_multiclass'])
            self.assertFalse(props['handles_multilabel'])

    def test_find_regressors(self):
        regressors = regression_components._regressors
        self.assertGreaterEqual(len(regressors), 1)
        for key in regressors:
            if hasattr(regressors[key], 'get_components'):
                continue
            self.assertIn(AutoSklearnRegressionAlgorithm,
                            regressors[key].__bases__)

    def test_find_preprocessors(self):
        preprocessors = preprocessing_components._preprocessors
        self.assertGreaterEqual(len(preprocessors),  1)
        for key in preprocessors:
            if hasattr(preprocessors[key], 'get_components'):
                continue
            self.assertIn(AutoSklearnPreprocessingAlgorithm,
                            preprocessors[key].__bases__)

    def test_configurations(self):
        cs = SimpleRegressionPipeline().get_hyperparameter_search_space()

        self._test_configurations(cs)

    def test_configurations_signed_data(self):
        dataset_properties = {'signed': True}
        cs = SimpleRegressionPipeline(dataset_properties=dataset_properties).\
            get_hyperparameter_search_space()

        self._test_configurations(configurations_space=cs,
                                  dataset_properties=dataset_properties)

    def test_configurations_sparse(self):
        dataset_properties = {'sparse': True}
        cs = SimpleRegressionPipeline(
            dataset_properties=dataset_properties
        ).get_hyperparameter_search_space()

        self._test_configurations(cs, make_sparse=True,
                                  dataset_properties=dataset_properties)

    def _test_configurations(self, configurations_space, make_sparse=False,
                             data=None, dataset_properties=None):
        # Use a limit of ~4GiB
        limit = 3072 * 1024 * 1024
        resource.setrlimit(resource.RLIMIT_AS, (limit, limit))

        configurations_space.seed(1)

        for i in range(10):
            config = configurations_space.sample_configuration()
            config._populate_values()

            # Restrict configurations which could take too long on travis-ci
            restrictions = {'regressor:adaboost:n_estimators': 50,
                            'regressor:adaboost:max_depth': 1,
                            'preprocessor:kernel_pca:n_components': 10,
                            'preprocessor:kitchen_sinks:n_components': 50,
                            'regressor:libsvm_svc:degree': 2,
                            'regressor:libsvm_svr:degree': 2,
                            'preprocessor:truncatedSVD:target_dim': 10,
                            'preprocessor:polynomial:degree': 2,
                            'regressor:lda:n_components': 10}

            for restrict_parameter in restrictions:
                restrict_to = restrictions[restrict_parameter]
                if restrict_parameter in config and \
                                config[restrict_parameter] is not None:
                    config._values[restrict_parameter] = restrict_to

            if data is None:
                X_train, Y_train, X_test, Y_test = get_dataset(
                    dataset='boston', make_sparse=make_sparse, add_NaNs=True)
            else:
                X_train = data['X_train'].copy()
                Y_train = data['Y_train'].copy()
                X_test = data['X_test'].copy()
                Y_test = data['Y_test'].copy()

            cls = SimpleRegressionPipeline(random_state=1,
                                           dataset_properties=dataset_properties)
            cls.set_hyperparameters(config)
            try:
                cls.fit(X_train, Y_train)
                predictions = cls.predict(X_test)
            except MemoryError as e:
                continue
            except ValueError as e:
                if "Floating-point under-/overflow occurred at epoch" in \
                        e.args[0]:
                    continue
                elif "removed all features" in e.args[0]:
                    continue
                elif "all features are discarded" in e.args[0]:
                    continue
                elif "Numerical problems in QDA" in e.args[0]:
                    continue
                elif 'Bug in scikit-learn' in e.args[0]:
                    continue
                elif 'The condensed distance matrix must contain only finite ' \
                     'values.' in e.args[0]:
                    continue
                else:
                    print(config)
                    print(traceback.format_exc())
                    raise e
            except RuntimeWarning as e:
                if "invalid value encountered in sqrt" in e.args[0]:
                    continue
                elif "divide by zero encountered in" in e.args[0]:
                    continue
                elif "invalid value encountered in divide" in e.args[0]:
                    continue
                elif "invalid value encountered in true_divide" in e.args[0]:
                    continue
                else:
                    print(config)
                    traceback.print_tb(sys.exc_info()[2])
                    raise e
            except UserWarning as e:
                if "FastICA did not converge" in e.args[0]:
                    continue
                else:
                    print(config)
                    traceback.print_tb(sys.exc_info()[2])
                    raise e
            except Exception as e:
                if "Multiple input features cannot have the same target value" in e.args[0]:
                    continue
                else:
                    print(config)
                    traceback.print_tb(sys.exc_info()[2])
                    raise e

    def test_default_configuration(self):
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(dataset='diabetes')
            auto = SimpleRegressionPipeline()
            auto = auto.fit(X_train, Y_train)
            predictions = auto.predict(copy.deepcopy(X_test))
            # The lower the worse
            r2_score = sklearn.metrics.r2_score(Y_test, predictions)
            self.assertAlmostEqual(0.339, r2_score, places=3)
            model_score = auto.score(copy.deepcopy(X_test), Y_test)
            self.assertAlmostEqual(model_score, r2_score, places=5)

    def test_default_configuration_iterative_fit(self):
        regressor = SimpleRegressionPipeline(
            include={'regressor': ['random_forest'],
                     'preprocessor': ['no_preprocessing']})
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='boston')
        XT = regressor.fit_transformer(X_train, Y_train)
        for i in range(1, 11):
            regressor.iterative_fit(X_train, Y_train)
            self.assertEqual(regressor.steps[-1][-1].choice.estimator.n_estimators,
                             i)

    def test_repr(self):
        representation = repr(SimpleRegressionPipeline())
        cls = eval(representation)
        self.assertIsInstance(cls, SimpleRegressionPipeline)

    def test_get_hyperparameter_search_space(self):
        cs = SimpleRegressionPipeline().get_hyperparameter_search_space()
        self.assertIsInstance(cs, ConfigurationSpace)
        conditions = cs.get_conditions()
        hyperparameters = cs.get_hyperparameters()
        self.assertEqual(158, len(hyperparameters))
        self.assertEqual(len(hyperparameters) - 5, len(conditions))

    def test_get_hyperparameter_search_space_include_exclude_models(self):
        cs = SimpleRegressionPipeline(
            include={'regressor': ['random_forest']}).get_hyperparameter_search_space()
        self.assertEqual(cs.get_hyperparameter('regressor:__choice__'),
            CategoricalHyperparameter('regressor:__choice__', ['random_forest']))

        # TODO add this test when more than one regressor is present
        cs = SimpleRegressionPipeline(exclude={'regressor': ['random_forest']}).\
            get_hyperparameter_search_space()
        self.assertNotIn('random_forest', str(cs))

        cs = SimpleRegressionPipeline(include={'preprocessor': ['pca']}).\
            get_hyperparameter_search_space()
        self.assertEqual(cs.get_hyperparameter('preprocessor:__choice__'),
            CategoricalHyperparameter('preprocessor:__choice__', ['pca']))

        cs = SimpleRegressionPipeline(exclude={'preprocessor': ['no_preprocessing']}).\
            get_hyperparameter_search_space()
        self.assertNotIn('no_preprocessing', str(cs))

    def test_get_hyperparameter_search_space_preprocessor_contradicts_default_classifier(
            self):
        cs = SimpleRegressionPipeline(include={'preprocessor': ['densifier']},
                                      dataset_properties={'sparse': True}).\
            get_hyperparameter_search_space()
        self.assertEqual(
            cs.get_hyperparameter('regressor:__choice__').default_value,
            'gradient_boosting'
        )

        cs = SimpleRegressionPipeline(include={'preprocessor': ['nystroem_sampler']}).\
            get_hyperparameter_search_space()
        self.assertEqual(
            cs.get_hyperparameter('regressor:__choice__').default_value,
            'sgd'
        )

    def test_get_hyperparameter_search_space_only_forbidden_combinations(self):
        self.assertRaisesRegexp(ValueError, "Cannot find a legal default "
                                            "configuration.",
                                SimpleRegressionPipeline,
                                include={'regressor': ['random_forest'],
                                         'preprocessor': ['kitchen_sinks']})

        # It must also be catched that no classifiers which can handle sparse
        # data are located behind the densifier
        self.assertRaisesRegexp(ValueError, "Cannot find a legal default "
                                            "configuration",
                                SimpleRegressionPipeline,
                                include={'regressor': ['ridge_regression'],
                                         'preprocessor': ['densifier']},
                                dataset_properties={'sparse': True})

    @unittest.skip("test_get_hyperparameter_search_space_dataset_properties" +
                   " Not yet Implemented")
    def test_get_hyperparameter_search_space_dataset_properties(self):
        # TODO: We do not have any dataset properties for regression, so this
        # test is somewhat stupid
        pass
        """
        full_cs = SimpleRegressionPipeline.get_hyperparameter_search_space()
        cs_mc = SimpleRegressionPipeline.get_hyperparameter_search_space()
        self.assertEqual(full_cs, cs_mc)

        cs_ml = SimpleRegressionPipeline.get_hyperparameter_search_space()
        self.assertNotIn('k_nearest_neighbors', str(cs_ml))
        self.assertNotIn('liblinear', str(cs_ml))
        self.assertNotIn('libsvm_svc', str(cs_ml))
        self.assertNotIn('sgd', str(cs_ml))

        cs_sp = SimpleRegressionPipeline.get_hyperparameter_search_space(
            sparse=True)
        self.assertNotIn('extra_trees', str(cs_sp))
        self.assertNotIn('gradient_boosting', str(cs_sp))
        self.assertNotIn('random_forest', str(cs_sp))

        cs_mc_ml = SimpleRegressionPipeline.get_hyperparameter_search_space()
        self.assertEqual(cs_ml, cs_mc_ml)

        self.assertRaisesRegexp(ValueError,
                                "No regressor to build a configuration space "
                                "for...", SimpleRegressionPipeline.
                                get_hyperparameter_search_space,
                                multiclass=True, multilabel=True, sparse=True)
    """

    def test_predict_batched(self):
        include = {'regressor': ['decision_tree']}
        cs = SimpleRegressionPipeline(include=include).get_hyperparameter_search_space()
        default = cs.get_default_configuration()
        regressor = SimpleRegressionPipeline(default, include=include)

        X_train, Y_train, X_test, Y_test = get_dataset(dataset='boston')
        regressor.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = regressor.predict(X_test_)
        mock_predict = unittest.mock.Mock(wraps=regressor.steps[-1][-1].predict)
        regressor.steps[-1][-1].predict = mock_predict
        prediction = regressor.predict(X_test, batch_size=20)
        self.assertEqual((356,), prediction.shape)
        self.assertEqual(18, mock_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

    def test_predict_batched_sparse(self):
        dataset_properties = {'sparse': True}
        include = {'regressor': ['decision_tree']}
        cs = SimpleRegressionPipeline(dataset_properties=dataset_properties,
                                      include=include).get_hyperparameter_search_space()
        default = cs.get_default_configuration()
        regressor = SimpleRegressionPipeline(default,
                                             dataset_properties=dataset_properties,
                                             include=include)

        X_train, Y_train, X_test, Y_test = get_dataset(dataset='boston',
                                                       make_sparse=True)
        regressor.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = regressor.predict(X_test_)
        mock_predict = unittest.mock.Mock(wraps=regressor.steps[-1][-1].predict)
        regressor.steps[-1][-1].predict = mock_predict
        prediction = regressor.predict(X_test, batch_size=20)
        self.assertEqual((356,), prediction.shape)
        self.assertEqual(18, mock_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

    @unittest.skip("test_check_random_state Not yet Implemented")
    def test_check_random_state(self):
        raise NotImplementedError()

    @unittest.skip("test_validate_input_X Not yet Implemented")
    def test_validate_input_X(self):
        raise NotImplementedError()

    @unittest.skip("test_validate_input_Y Not yet Implemented")
    def test_validate_input_Y(self):
        raise NotImplementedError()

    def test_set_params(self):
        pass

    def test_get_params(self):
        pass



import pickle
import unittest

from autosklearn.util import get_logger


class TestLogging(unittest.TestCase):

    def test_pickle_logger(self):
        logger = get_logger('Test')
        p = pickle.dumps(logger, protocol=-1)


import unittest

import numpy as np
from scipy import sparse

from sklearn.utils.testing import assert_equal
from sklearn.utils.testing import assert_array_equal
from sklearn.utils.testing import assert_raises
from sklearn.utils.testing import assert_false
from sklearn.utils.testing import assert_true

from sklearn.preprocessing import Imputer
from sklearn.pipeline import Pipeline
from sklearn.model_selection import GridSearchCV
from sklearn import tree
from sklearn.random_projection import sparse_random_matrix


def safe_median(arr, *args, **kwargs):
    # np.median([]) raises a TypeError for numpy >= 1.10.1
    length = arr.size if hasattr(arr, 'size') else len(arr)
    return np.nan if length == 0 else np.median(arr, *args, **kwargs)


def safe_mean(arr, *args, **kwargs):
    # np.mean([]) raises a RuntimeWarning for numpy >= 1.10.1
    length = arr.size if hasattr(arr, 'size') else len(arr)
    return np.nan if length == 0 else np.mean(arr, *args, **kwargs)


class ImputationTest(unittest.TestCase):
    def _check_statistics(self, X, X_true,
                          strategy, statistics, missing_values):
        """Utility function for testing imputation for a given strategy.

        Test:
            - along the two axes
            - with dense and sparse arrays

        Check that:
            - the statistics (mean, median, mode) are correct
            - the missing values are imputed correctly"""

        err_msg = "Parameters: strategy = %s, missing_values = %s, " \
                  "axis = {0}, sparse = {1}" % (strategy, missing_values)

        # Normal matrix, axis = 0
        imputer = Imputer(missing_values, strategy=strategy, axis=0)
        X_trans = imputer.fit(X).transform(X.copy())
        assert_array_equal(imputer.statistics_, statistics,
                           err_msg.format(0, False))
        assert_array_equal(X_trans, X_true, err_msg.format(0, False))

        # Normal matrix, axis = 1
        imputer = Imputer(missing_values, strategy=strategy, axis=1)
        imputer.fit(X.transpose())
        if np.isnan(statistics).any():
            assert_raises(ValueError, imputer.transform, X.copy().transpose())
        else:
            X_trans = imputer.transform(X.copy().transpose())
            assert_array_equal(X_trans, X_true.transpose(),
                               err_msg.format(1, False))

        # Sparse matrix, axis = 0
        imputer = Imputer(missing_values, strategy=strategy, axis=0)
        imputer.fit(sparse.csc_matrix(X))
        X_trans = imputer.transform(sparse.csc_matrix(X.copy()))

        if sparse.issparse(X_trans):
            X_trans = X_trans.toarray()

        assert_array_equal(imputer.statistics_, statistics,
                           err_msg.format(0, True))
        assert_array_equal(X_trans, X_true, err_msg.format(0, True))

        # Sparse matrix, axis = 1
        imputer = Imputer(missing_values, strategy=strategy, axis=1)
        imputer.fit(sparse.csc_matrix(X.transpose()))
        if np.isnan(statistics).any():
            assert_raises(ValueError, imputer.transform,
                          sparse.csc_matrix(X.copy().transpose()))
        else:
            X_trans = imputer.transform(sparse.csc_matrix(X.copy().transpose()))

            if sparse.issparse(X_trans):
                X_trans = X_trans.toarray()

            assert_array_equal(X_trans, X_true.transpose(),
                               err_msg.format(1, True))


    def test_imputation_shape(self):
        """Verify the shapes of the imputed matrix for different strategies."""
        X = np.random.randn(10, 2)
        X[::2] = np.nan

        for strategy in ['mean', 'median', 'most_frequent']:
            imputer = Imputer(strategy=strategy)
            X_imputed = imputer.fit_transform(X)
            assert_equal(X_imputed.shape, (10, 2))
            X_imputed = imputer.fit_transform(sparse.csr_matrix(X))
            assert_equal(X_imputed.shape, (10, 2))


    def test_imputation_mean_median_only_zero(self):
        """Test imputation using the mean and median strategies, when
           missing_values == 0."""
        X = np.array([
            [np.nan, 0, 0, 0, 5],
            [np.nan, 1, 0, np.nan, 3],
            [np.nan, 2, 0, 0, 0],
            [np.nan, 6, 0, 5, 13],
        ])

        X_imputed_mean = np.array([
            [3, 5],
            [1, 3],
            [2, 7],
            [6, 13],
        ])
        statistics_mean = [np.nan, 3, np.nan, np.nan, 7]

        # Behaviour of median with NaN is undefined, e.g. different results in
        # np.median and np.ma.median
        X_for_median = X[:, [0, 1, 2, 4]]
        X_imputed_median = np.array([
            [2, 5],
            [1, 3],
            [2, 5],
            [6, 13],
        ])
        statistics_median = [np.nan, 2, np.nan, 5]

        self._check_statistics(X, X_imputed_mean, "mean", statistics_mean, 0)
        self._check_statistics(X_for_median, X_imputed_median, "median",
                          statistics_median, 0)


    def test_imputation_mean_median(self):
        """Test imputation using the mean and median strategies, when
           missing_values != 0."""
        rng = np.random.RandomState(0)

        dim = 10
        dec = 10
        shape = (dim * dim, dim + dec)

        zeros = np.zeros(shape[0])
        values = np.arange(1, shape[0] + 1)
        values[4::2] = - values[4::2]

        tests = [("mean", "NaN", lambda z, v, p: safe_mean(np.hstack((z, v)))),
                 ("mean", 0, lambda z, v, p: np.mean(v)),
                 ("median", "NaN", lambda z, v, p: safe_median(np.hstack((z, v)))),
                 ("median", 0, lambda z, v, p: np.median(v))]

        for strategy, test_missing_values, true_value_fun in tests:
            X = np.empty(shape)
            X_true = np.empty(shape)
            true_statistics = np.empty(shape[1])

            # Create a matrix X with columns
            # - with only zeros,
            #    - with only missing values
            #    - with zeros, missing values and values
            # And a matrix X_true containing all true values
            for j in range(shape[1]):
                nb_zeros = (j - dec + 1 > 0) * (j - dec + 1) * (j - dec + 1)
                nb_missing_values = max(shape[0] + dec * dec
                                        - (j + dec) * (j + dec), 0)
                nb_values = shape[0] - nb_zeros - nb_missing_values

                z = zeros[:nb_zeros]
                p = np.repeat(test_missing_values, nb_missing_values)
                v = values[rng.permutation(len(values))[:nb_values]]
                true_statistics[j] = true_value_fun(z, v, p)

                # Create the columns
                X[:, j] = np.hstack((v, z, p))

                if 0 == test_missing_values:
                    X_true[:, j] = np.hstack((v,
                                              np.repeat(
                                                  true_statistics[j],
                                                  nb_missing_values + nb_zeros)))
                else:
                    X_true[:, j] = np.hstack((v,
                                              z,
                                              np.repeat(true_statistics[j],
                                                        nb_missing_values)))

                # Shuffle them the same way
                np.random.RandomState(j).shuffle(X[:, j])
                np.random.RandomState(j).shuffle(X_true[:, j])

            # Mean doesn't support columns containing NaNs, median does
            if strategy == "median":
                cols_to_keep = ~np.isnan(X_true).any(axis=0)
            else:
                cols_to_keep = ~np.isnan(X_true).all(axis=0)

            X_true = X_true[:, cols_to_keep]

            self._check_statistics(X, X_true, strategy,
                              true_statistics, test_missing_values)


    def test_imputation_median_special_cases(self):
        """Test median imputation with sparse boundary cases
        """
        X = np.array([
            [0, np.nan, np.nan],  # odd: implicit zero
            [5, np.nan, np.nan],  # odd: explicit nonzero
            [0, 0, np.nan],  # even: average two zeros
            [-5, 0, np.nan],  # even: avg zero and neg
            [0, 5, np.nan],  # even: avg zero and pos
            [4, 5, np.nan],  # even: avg nonzeros
            [-4, -5, np.nan],  # even: avg negatives
            [-1, 2, np.nan],  # even: crossing neg and pos
        ]).transpose()

        X_imputed_median = np.array([
            [0, 0, 0],
            [5, 5, 5],
            [0, 0, 0],
            [-5, 0, -2.5],
            [0, 5, 2.5],
            [4, 5, 4.5],
            [-4, -5, -4.5],
            [-1, 2, .5],
        ]).transpose()
        statistics_median = [0, 5, 0, -2.5, 2.5, 4.5, -4.5, .5]

        self._check_statistics(X, X_imputed_median, "median",
                          statistics_median, 'NaN')


    def test_imputation_most_frequent(self):
        """Test imputation using the most-frequent strategy."""
        X = np.array([
            [-1, -1, 0, 5],
            [-1, 2, -1, 3],
            [-1, 1, 3, -1],
            [-1, 2, 3, 7],
        ])

        X_true = np.array([
            [2, 0, 5],
            [2, 3, 3],
            [1, 3, 3],
            [2, 3, 7],
        ])

        # scipy.stats.mode, used in Imputer, doesn't return the first most
        # frequent as promised in the doc but the lowest most frequent. When this
        # test will fail after an update of scipy, Imputer will need to be updated
        # to be consistent with the new (correct) behaviour
        self._check_statistics(X, X_true, "most_frequent", [np.nan, 2, 3, 3],
                               -1)


    def test_imputation_pipeline_grid_search(self):
        """Test imputation within a pipeline + gridsearch."""
        pipeline = Pipeline([('imputer', Imputer(missing_values=0)),
                             ('tree', tree.DecisionTreeRegressor(random_state=0))])

        parameters = {
            'imputer__strategy': ["mean", "median", "most_frequent"],
            'imputer__axis': [0, 1]
        }

        l = 100
        X = sparse_random_matrix(l, l, density=0.10)
        Y = sparse_random_matrix(l, 1, density=0.10).toarray()
        gs = GridSearchCV(pipeline, parameters)
        gs.fit(X, Y)


    def test_imputation_pickle(self):
        """Test for pickling imputers."""
        import pickle

        l = 100
        X = sparse_random_matrix(l, l, density=0.10)

        for strategy in ["mean", "median", "most_frequent"]:
            imputer = Imputer(missing_values=0, strategy=strategy)
            imputer.fit(X)

            imputer_pickled = pickle.loads(pickle.dumps(imputer))

            assert_array_equal(imputer.transform(X.copy()),
                               imputer_pickled.transform(X.copy()),
                               "Fail to transform the data after pickling "
                               "(strategy = %s)" % (strategy))


    def test_imputation_copy(self):
        """Test imputation with copy"""
        X_orig = sparse_random_matrix(5, 5, density=0.75, random_state=0)

        # copy=True, dense => copy
        X = X_orig.copy().toarray()
        imputer = Imputer(missing_values=0, strategy="mean", copy=True)
        Xt = imputer.fit(X).transform(X)
        Xt[0, 0] = -1
        assert_false(np.all(X == Xt))

        # copy=True, sparse csr => copy
        X = X_orig.copy()
        imputer = Imputer(missing_values=X.data[0], strategy="mean", copy=True)
        Xt = imputer.fit(X).transform(X)
        Xt.data[0] = -1
        assert_false(np.all(X.data == Xt.data))

        # copy=False, dense => no copy
        X = X_orig.copy().toarray()
        imputer = Imputer(missing_values=0, strategy="mean", copy=False)
        Xt = imputer.fit(X).transform(X)
        Xt[0, 0] = -1
        assert_true(np.all(X == Xt))

        # copy=False, sparse csr, axis=1 => no copy
        X = X_orig.copy()
        imputer = Imputer(missing_values=X.data[0], strategy="mean",
                          copy=False, axis=1)
        Xt = imputer.fit(X).transform(X)
        Xt.data[0] = -1
        assert_true(np.all(X.data == Xt.data))

        # copy=False, sparse csc, axis=0 => no copy
        X = X_orig.copy().tocsc()
        imputer = Imputer(missing_values=X.data[0], strategy="mean",
                          copy=False, axis=0)
        Xt = imputer.fit(X).transform(X)
        Xt.data[0] = -1
        assert_true(np.all(X.data == Xt.data))

        # copy=False, sparse csr, axis=0 => copy
        X = X_orig.copy()
        imputer = Imputer(missing_values=X.data[0], strategy="mean",
                          copy=False, axis=0)
        Xt = imputer.fit(X).transform(X)
        Xt.data[0] = -1
        assert_false(np.all(X.data == Xt.data))

        # copy=False, sparse csc, axis=1 => copy
        X = X_orig.copy().tocsc()
        imputer = Imputer(missing_values=X.data[0], strategy="mean",
                          copy=False, axis=1)
        Xt = imputer.fit(X).transform(X)
        Xt.data[0] = -1
        assert_false(np.all(X.data == Xt.data))

        # copy=False, sparse csr, axis=1, missing_values=0 => copy
        X = X_orig.copy()
        imputer = Imputer(missing_values=0, strategy="mean",
                          copy=False, axis=1)
        Xt = imputer.fit(X).transform(X)
        assert_false(sparse.issparse(Xt))

        # Note: If X is sparse and if missing_values=0, then a (dense) copy of X is
        # made, even if copy=False.



# -*- encoding: utf-8 -*-
import setuptools
from setuptools.extension import Extension
import numpy as np
from Cython.Build import cythonize
import os
import sys


# Check if Auto-sklearn *could* run on the given system
if os.name != 'posix':
    raise ValueError(
        'Detected unsupported operating system: %s. Please check '
        'the compability information of auto-sklearn: http://automl.github.io'
        '/auto-sklearn/stable/installation.html#windows-osx-compability' %
        sys.platform
    )

if sys.version_info < (3, 5):
    raise ValueError(
        'Unsupported python version %s found. Auto-sklearn requires Python '
        '3.5 or higher.' % sys.version_info
    )


extensions = cythonize(
    [Extension('autosklearn.data.competition_c_functions',
               sources=['autosklearn/data/competition_c_functions.pyx'],
               language='c',
               include_dirs=[np.get_include()])
     ])

requirements = [
    "setuptools",
    "nose",
    "six",
    "Cython",
    "numpy>=1.9.0",
    "scipy>=0.14.1",
    "scikit-learn>=0.19,<0.20",
    "lockfile",
    "joblib",
    "psutil",
    "pyyaml",
    "liac-arff",
    "pandas",
    "ConfigSpace>=0.4.0,<0.5",
    "pynisher>=0.4,<0.5",
    "pyrfr>=0.6.1,<0.8",
    "smac>=0.8,<0.9",
    "xgboost==0.7.post3",
]

with open("autosklearn/__version__.py") as fh:
    version = fh.readlines()[-1].split()[-1].strip("\"'")

setuptools.setup(
    name='auto-sklearn',
    description='Automated machine learning.',
    version=version,
    ext_modules=extensions,
    packages=setuptools.find_packages(exclude=['test']),
    install_requires=requirements,
    test_suite='nose.collector',
    include_package_data=True,
    author='Matthias Feurer',
    author_email='feurerm@informatik.uni-freiburg.de',
    license='BSD',
    platforms=['Linux'],
    classifiers=[],
    python_requires='>=3.4.*',
    url='https://automl.github.io/auto-sklearn',
)



import copy
import os
import resource
import tempfile
import traceback
import unittest
import unittest.mock

import numpy as np
import sklearn.datasets
import sklearn.decomposition
import sklearn.model_selection
import sklearn.ensemble
import sklearn.svm
from sklearn.utils.testing import assert_array_almost_equal
from sklearn.externals.joblib import Memory

from ConfigSpace.configuration_space import ConfigurationSpace, \
    Configuration
from ConfigSpace.hyperparameters import CategoricalHyperparameter

from autosklearn.pipeline.classification import SimpleClassificationPipeline
from autosklearn.pipeline.components.base import \
    AutoSklearnClassificationAlgorithm, AutoSklearnPreprocessingAlgorithm
import autosklearn.pipeline.components.classification as classification_components
import autosklearn.pipeline.components.feature_preprocessing as preprocessing_components
from autosklearn.pipeline.util import get_dataset
from autosklearn.pipeline.constants import *


class DummyClassifier(AutoSklearnClassificationAlgorithm):
    @staticmethod
    def get_properties(dataset_properties=None):
        return {'shortname': 'AB',
                'name': 'AdaBoost Classifier',
                'handles_regression': False,
                'handles_classification': True,
                'handles_multiclass': True,
                'handles_multilabel': True,
                'is_deterministic': True,
                'input': (DENSE, SPARSE, UNSIGNED_DATA),
                'output': (PREDICTIONS,)}

    @staticmethod
    def get_hyperparameter_search_space(dataset_properties=None):
        cs = ConfigurationSpace()
        return cs


class DummyPreprocessor(AutoSklearnPreprocessingAlgorithm):
    @staticmethod
    def get_properties(dataset_properties=None):
        return {'shortname': 'AB',
                'name': 'AdaBoost Classifier',
                'handles_regression': False,
                'handles_classification': True,
                'handles_multiclass': True,
                'handles_multilabel': True,
                'is_deterministic': True,
                'input': (DENSE, SPARSE, UNSIGNED_DATA),
                'output': (INPUT,)}

    @staticmethod
    def get_hyperparameter_search_space(dataset_properties=None):
        cs = ConfigurationSpace()
        return cs


class SimpleClassificationPipelineTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_io_dict(self):
        classifiers = classification_components._classifiers
        for c in classifiers:
            if classifiers[c] == classification_components.ClassifierChoice:
                continue
            props = classifiers[c].get_properties()
            self.assertIn('input', props)
            self.assertIn('output', props)
            inp = props['input']
            output = props['output']

            self.assertIsInstance(inp, tuple)
            self.assertIsInstance(output, tuple)
            for i in inp:
                self.assertIn(i, (SPARSE, DENSE, SIGNED_DATA, UNSIGNED_DATA))
            self.assertEqual(output, (PREDICTIONS,))
            self.assertIn('handles_regression', props)
            self.assertFalse(props['handles_regression'])
            self.assertIn('handles_classification', props)
            self.assertIn('handles_multiclass', props)
            self.assertIn('handles_multilabel', props)

    def test_find_classifiers(self):
        classifiers = classification_components._classifiers
        self.assertGreaterEqual(len(classifiers), 2)
        for key in classifiers:
            if hasattr(classifiers[key], 'get_components'):
                continue
            self.assertIn(AutoSklearnClassificationAlgorithm,
                            classifiers[key].__bases__)

    def test_find_preprocessors(self):
        preprocessors = preprocessing_components._preprocessors
        self.assertGreaterEqual(len(preprocessors),  1)
        for key in preprocessors:
            if hasattr(preprocessors[key], 'get_components'):
                continue
            self.assertIn(AutoSklearnPreprocessingAlgorithm,
                            preprocessors[key].__bases__)

    def test_default_configuration(self):
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(dataset='iris')
            auto = SimpleClassificationPipeline()
            auto = auto.fit(X_train, Y_train)
            predictions = auto.predict(X_test)
            self.assertAlmostEqual(0.94,
                sklearn.metrics.accuracy_score(predictions, Y_test))
            scores = auto.predict_proba(X_test)

    def test_default_configuration_multilabel(self):
        for i in range(2):
            dataset_properties = {'multilabel': True}
            classifier = SimpleClassificationPipeline(
                dataset_properties=dataset_properties)
            cs = classifier.get_hyperparameter_search_space()
            default = cs.get_default_configuration()
            X_train, Y_train, X_test, Y_test = get_dataset(dataset='iris',
                                                           make_multilabel=True)
            classifier.set_hyperparameters(default)
            classifier = classifier.fit(X_train, Y_train)
            predictions = classifier.predict(X_test)
            self.assertAlmostEqual(0.94,
                                   sklearn.metrics.accuracy_score(predictions,
                                                                  Y_test))
            scores = classifier.predict_proba(X_test)

    def test_default_configuration_iterative_fit(self):
        classifier = SimpleClassificationPipeline(
            include={'classifier': ['random_forest'],
                     'preprocessor': ['no_preprocessing']})
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='iris')
        XT = classifier.fit_transformer(X_train, Y_train)
        for i in range(1, 11):
            classifier.iterative_fit(X_train, Y_train)
            self.assertEqual(classifier.steps[-1][-1].choice.estimator.n_estimators,
                             i)

    def test_repr(self):
        representation = repr(SimpleClassificationPipeline())
        cls = eval(representation)
        self.assertIsInstance(cls, SimpleClassificationPipeline)

    def test_multilabel(self):
        cache = Memory(cachedir=tempfile.gettempdir())
        cached_func = cache.cache(
            sklearn.datasets.make_multilabel_classification
        )
        X, Y = cached_func(
            n_samples=150,
            n_features=20,
            n_classes=5,
            n_labels=2,
            length=50,
            allow_unlabeled=True,
            sparse=False,
            return_indicator=True,
            return_distributions=False,
            random_state=1
        )
        X_train = X[:100, :]
        Y_train = Y[:100, :]
        X_test = X[101:, :]
        Y_test = Y[101:, ]

        data = {'X_train': X_train, 'Y_train': Y_train,
                'X_test': X_test, 'Y_test': Y_test}

        dataset_properties = {'multilabel': True}
        cs = SimpleClassificationPipeline(dataset_properties=dataset_properties).\
            get_hyperparameter_search_space()
        self._test_configurations(configurations_space=cs, data=data)

    def test_configurations(self):
        cs = SimpleClassificationPipeline().get_hyperparameter_search_space()

        self._test_configurations(configurations_space=cs)

    def test_configurations_signed_data(self):
        dataset_properties = {'signed': True}
        cs = SimpleClassificationPipeline(dataset_properties=dataset_properties)\
            .get_hyperparameter_search_space()

        self._test_configurations(configurations_space=cs,
                                  dataset_properties=dataset_properties)

    def test_configurations_sparse(self):
        cs = SimpleClassificationPipeline(dataset_properties={'sparse': True}).\
            get_hyperparameter_search_space()

        self._test_configurations(configurations_space=cs, make_sparse=True)

    def test_configurations_categorical_data(self):
        cs = SimpleClassificationPipeline(
            dataset_properties={'sparse': False},
            include={
                'preprocessor': ['no_preprocessing'],
                'classifier': ['sgd', 'gradient_boosting']
            }
        ).get_hyperparameter_search_space()

        categorical = [True, True, True, False, False, True, True, True,
                       False, True, True, True, True, True, True, True,
                       True, True, True, True, True, True, True, True, True,
                       True, True, True, True, True, True, True, False,
                       False, False, True, True, True]
        this_directory = os.path.dirname(__file__)
        X = np.loadtxt(os.path.join(this_directory, "components",
                                    "data_preprocessing", "dataset.pkl"))
        y = X[:, -1].copy()
        X = X[:,:-1]
        X_train, X_test, Y_train, Y_test = \
            sklearn.model_selection.train_test_split(X, y)
        data = {'X_train': X_train, 'Y_train': Y_train,
                'X_test': X_test, 'Y_test': Y_test}

        init_params = {
            'categorical_encoding:one_hot_encoding:categorical_features':
                categorical
        }

        self._test_configurations(configurations_space=cs, make_sparse=True,
                                  data=data, init_params=init_params)

    @unittest.mock.patch('autosklearn.pipeline.components.data_preprocessing'
                         '.one_hot_encoding.OHEChoice.set_hyperparameters')
    def test_categorical_passed_to_one_hot_encoder(self, ohe_mock):
        cls = SimpleClassificationPipeline(
            init_params={
                'categorical_encoding:one_hot_encoding:categorical_features':
                    [True, False]
            }
        )
        self.assertEqual(
            ohe_mock.call_args[1]['init_params'],
            {'one_hot_encoding:categorical_features': [True, False]}
        )
        default = cls.get_hyperparameter_search_space().get_default_configuration()
        cls.set_hyperparameters(configuration=default,
            init_params={
                'categorical_encoding:one_hot_encoding:categorical_features':
                    [True, True, False]
            }
        )
        self.assertEqual(
            ohe_mock.call_args[1]['init_params'],
            {'one_hot_encoding:categorical_features': [True, True, False]}
        )

    def _test_configurations(self, configurations_space, make_sparse=False,
                             data=None, init_params=None,
                             dataset_properties=None):
        # Use a limit of ~3GiB
        limit = 3072 * 1024 * 1024
        resource.setrlimit(resource.RLIMIT_AS, (limit, limit))

        print(configurations_space)

        for i in range(10):
            config = configurations_space.sample_configuration()
            config._populate_values()

            # Restrict configurations which could take too long on travis-ci
            restrictions = {'classifier:passive_aggressive:n_iter': 5,
                            'classifier:sgd:n_iter': 5,
                            'classifier:adaboost:n_estimators': 50,
                            'classifier:adaboost:max_depth': 1,
                            'preprocessor:kernel_pca:n_components': 10,
                            'preprocessor:kitchen_sinks:n_components': 50,
                            'classifier:proj_logit:max_epochs': 1,
                            'classifier:libsvm_svc:degree': 2,
                            'regressor:libsvm_svr:degree': 2,
                            'preprocessor:truncatedSVD:target_dim': 10,
                            'preprocessor:polynomial:degree': 2,
                            'classifier:lda:n_components': 10,
                            'preprocessor:nystroem_sampler:n_components': 50,
                            'preprocessor:feature_agglomeration:n_clusters': 2,
                            'classifier:gradient_boosting:max_depth': 2,
                            'classifier:gradient_boosting:n_estimators': 50}

            for restrict_parameter in restrictions:
                restrict_to = restrictions[restrict_parameter]
                if restrict_parameter in config and \
                        config[restrict_parameter] is not None:
                    config._values[restrict_parameter] = restrict_to

            print(config)

            if data is None:
                X_train, Y_train, X_test, Y_test = get_dataset(
                    dataset='digits', make_sparse=make_sparse, add_NaNs=True)
            else:
                X_train = data['X_train'].copy()
                Y_train = data['Y_train'].copy()
                X_test = data['X_test'].copy()
                Y_test = data['Y_test'].copy()

            init_params_ = copy.deepcopy(init_params)
            cls = SimpleClassificationPipeline(random_state=1,
                                               dataset_properties=dataset_properties,
                                               init_params=init_params_,)
            cls.set_hyperparameters(config, init_params=init_params_)
            try:
                cls.fit(X_train, Y_train, )
                predictions = cls.predict(X_test.copy())
                predictions = cls.predict_proba(X_test)
            except MemoryError as e:
                continue
            except ValueError as e:
                if "Floating-point under-/overflow occurred at epoch" in \
                        e.args[0]:
                    continue
                elif "removed all features" in e.args[0]:
                    continue
                elif "all features are discarded" in e.args[0]:
                    continue
                elif "Numerical problems in QDA" in e.args[0]:
                    continue
                elif 'Bug in scikit-learn' in e.args[0]:
                    continue
                elif 'The condensed distance matrix must contain only finite ' \
                     'values.' in e.args[0]:
                    continue
                else:
                    print(config)
                    print(traceback.format_exc())
                    raise e
            except RuntimeWarning as e:
                if "invalid value encountered in sqrt" in e.args[0]:
                    continue
                elif "divide by zero encountered in" in e.args[0]:
                    continue
                elif "invalid value encountered in divide" in e.args[0]:
                    continue
                elif "invalid value encountered in true_divide" in e.args[0]:
                    continue
                else:
                    print(traceback.format_exc())
                    print(config)
                    raise e
            except UserWarning as e:
                if "FastICA did not converge" in e.args[0]:
                    continue
                else:
                    print(traceback.format_exc())
                    print(config)
                    raise e
            #except XGBoostError as e:
            #    if "std::bad_alloc" in e.args[0]:
            #        continue
            #    else:
            #        print(traceback.format_exc())
            #        print(config)
            #        raise e

    def test_get_hyperparameter_search_space(self):
        cs = SimpleClassificationPipeline().get_hyperparameter_search_space()
        self.assertIsInstance(cs, ConfigurationSpace)
        conditions = cs.get_conditions()

        self.assertEqual(len(cs.get_hyperparameter(
            'rescaling:__choice__').choices), 6)
        self.assertEqual(len(cs.get_hyperparameter(
            'classifier:__choice__').choices), 16)
        self.assertEqual(len(cs.get_hyperparameter(
            'preprocessor:__choice__').choices), 13)

        hyperparameters = cs.get_hyperparameters()
        self.assertEqual(172, len(hyperparameters))

        #for hp in sorted([str(h) for h in hyperparameters]):
        #    print hp

        # The four parameters which are always active are classifier,
        # preprocessor, imputation strategy and scaling strategy
        self.assertEqual(len(hyperparameters) - 6, len(conditions))

    def test_get_hyperparameter_search_space_include_exclude_models(self):
        cs = SimpleClassificationPipeline(include={'classifier': ['libsvm_svc']})\
            .get_hyperparameter_search_space()
        self.assertEqual(cs.get_hyperparameter('classifier:__choice__'),
            CategoricalHyperparameter('classifier:__choice__', ['libsvm_svc']))

        cs = SimpleClassificationPipeline(exclude={'classifier': ['libsvm_svc']}).\
            get_hyperparameter_search_space()
        self.assertNotIn('libsvm_svc', str(cs))

        cs = SimpleClassificationPipeline(
            include={'preprocessor': ['select_percentile_classification']}).\
            get_hyperparameter_search_space()
        self.assertEqual(cs.get_hyperparameter('preprocessor:__choice__'),
            CategoricalHyperparameter('preprocessor:__choice__',
                                      ['select_percentile_classification']))

        cs = SimpleClassificationPipeline(
            exclude={'preprocessor': ['select_percentile_classification']}
        ).get_hyperparameter_search_space()
        self.assertNotIn('select_percentile_classification', str(cs))

    def test_get_hyperparameter_search_space_preprocessor_contradicts_default_classifier(self):
        cs = SimpleClassificationPipeline(
            include={'preprocessor': ['densifier']}, dataset_properties={'sparse': True}).\
            get_hyperparameter_search_space()
        self.assertEqual(cs.get_hyperparameter(
            'classifier:__choice__').default_value,
            'qda'
        )

        cs = SimpleClassificationPipeline(include={'preprocessor': ['nystroem_sampler']}).\
            get_hyperparameter_search_space()
        self.assertEqual(cs.get_hyperparameter(
            'classifier:__choice__').default_value,
            'sgd'
        )

    def test_get_hyperparameter_search_space_only_forbidden_combinations(self):
        self.assertRaisesRegexp(AssertionError, "No valid pipeline found.",
                                SimpleClassificationPipeline,
                                include={'classifier': ['multinomial_nb'],
                                         'preprocessor': ['pca']},
                                dataset_properties={'sparse': True})

        # It must also be catched that no classifiers which can handle sparse
        #  data are located behind the densifier
        self.assertRaisesRegexp(ValueError, "Cannot find a legal default "
                                            "configuration.",
                                SimpleClassificationPipeline,
                                include={'classifier': ['liblinear_svc'],
                                         'preprocessor': ['densifier']},
                                dataset_properties={'sparse': True})

    @unittest.skip("Wait until ConfigSpace is fixed.")
    def test_get_hyperparameter_search_space_dataset_properties(self):
        cs_mc = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={'multiclass': True})
        self.assertNotIn('bernoulli_nb', str(cs_mc))

        cs_ml = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={'multilabel': True})
        self.assertNotIn('k_nearest_neighbors', str(cs_ml))
        self.assertNotIn('liblinear', str(cs_ml))
        self.assertNotIn('libsvm_svc', str(cs_ml))
        self.assertNotIn('sgd', str(cs_ml))

        cs_sp = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={'sparse': True})
        self.assertIn('extra_trees', str(cs_sp))
        self.assertIn('gradient_boosting', str(cs_sp))
        self.assertIn('random_forest', str(cs_sp))

        cs_mc_ml = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={'multilabel': True, 'multiclass': True})
        self.assertEqual(cs_ml, cs_mc_ml)

    def test_predict_batched(self):
        cls = SimpleClassificationPipeline(include={'classifier': ['sgd']})

        # Multiclass
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='digits')
        cls.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)
        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict
        prediction = cls.predict_proba(X_test, batch_size=20)
        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

    def test_predict_batched_sparse(self):
        cls = SimpleClassificationPipeline(dataset_properties={'sparse': True},
                                           include={'classifier': ['sgd']})

        # Multiclass
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='digits',
                                                       make_sparse=True)
        cls.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)
        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict
        prediction = cls.predict_proba(X_test, batch_size=20)
        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

    def test_predict_proba_batched(self):
        # Multiclass
        cls = SimpleClassificationPipeline(include={'classifier': ['sgd']})
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='digits')

        cls.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)
        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict
        prediction = cls.predict_proba(X_test, batch_size=20)
        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

        # Multilabel
        cls = SimpleClassificationPipeline(include={'classifier': ['lda']})
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='digits')
        Y_train = np.array(list([(list([1 if i != y else 0 for i in range(10)]))
                                 for y in Y_train]))
        cls.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)
        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict
        prediction = cls.predict_proba(X_test, batch_size=20)
        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

    def test_predict_proba_batched_sparse(self):

        cls = SimpleClassificationPipeline(
            dataset_properties={'sparse': True, 'multiclass': True},
            include={'classifier': ['sgd']})

        # Multiclass
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='digits',
                                                       make_sparse=True)
        cls.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)
        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict
        prediction = cls.predict_proba(X_test, batch_size=20)
        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

        # Multilabel
        cls = SimpleClassificationPipeline(
            dataset_properties={'sparse': True, 'multilabel': True},
            include={'classifier': ['lda']})
        X_train, Y_train, X_test, Y_test = get_dataset(dataset='digits',
                                                       make_sparse=True)
        Y_train = np.array(list([(list([1 if i != y else 0 for i in range(10)]))
                                 for y in Y_train]))
        cls.fit(X_train, Y_train)
        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)
        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict
        prediction = cls.predict_proba(X_test, batch_size=20)
        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        assert_array_almost_equal(prediction_, prediction)

    def test_set_params(self):
        pass

    def test_get_params(self):
        pass

    def test_add_classifier(self):
        self.assertEqual(len(classification_components._addons.components), 0)
        classification_components.add_classifier(DummyClassifier)
        self.assertEqual(len(classification_components._addons.components), 1)
        cs = SimpleClassificationPipeline().get_hyperparameter_search_space()
        self.assertIn('DummyClassifier', str(cs))
        del classification_components._addons.components['DummyClassifier']

    def test_add_preprocessor(self):
        self.assertEqual(len(preprocessing_components._addons.components), 0)
        preprocessing_components.add_preprocessor(DummyPreprocessor)
        self.assertEqual(len(preprocessing_components._addons.components), 1)
        cs = SimpleClassificationPipeline().get_hyperparameter_search_space()
        self.assertIn('DummyPreprocessor', str(cs))
        del preprocessing_components._addons.components['DummyPreprocessor']




import sklearn.ensemble

from autosklearn.pipeline.components.classification.random_forest import \
    RandomForest

from .test_base import BaseClassificationComponentTest


class RandomForestComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.94
    res["iris_n_calls"] = 7
    res["default_iris_iterative"] = 0.95999999999999996
    res["default_iris_proba"] = 0.1357018315377785
    res["default_iris_sparse"] = 0.85999999999999999
    # TODO find source of discrepancy!
    res["default_digits"] = 0.899210686095932
    res["digits_n_calls"] = 7
    res["default_digits_iterative"] = 0.8828172434729812
    res["default_digits_binary"] = 0.9884638737097754
    res["default_digits_multilabel"] = 0.9965409490589346
    res["default_digits_multilabel_proba"] = 0.9922896819782431

    sk_mod = sklearn.ensemble.RandomForestClassifier
    module = RandomForest
    step_hyperparameter = {
        'name': 'n_estimators',
        'value': 100,
    }




from collections import OrderedDict
import copy
import cPickle
import glob
import itertools
import matplotlib.pyplot as plt
import matplotlib.colors
import numpy as np
import scipy.stats
import sys

import HPOlib.Plotting.plot_util as plot_util
import HPOlib.Plotting.statistics as statistics


def find_ground_truth(globstring):
    glob_results = glob.glob(globstring)
    if len(glob_results) > 1:
        raise Exception("There must be only one ground truth directory for "
                        "%s" % globstring)
    elif len(glob_results) == 0:
        print("Found no ground truth for %s" % globstring)
        return None

    with open(glob_results[0]) as fh:
        trials = cPickle.load(fh)

    return trials


def plot_rankings(trial_list, name_list, optimum=0, title="", log=False,
                  save="", y_min=0, y_max=0, cut=sys.maxint, figsize=(16, 6),
                  legend_ncols=4, bootstrap_samples=500):
    rankings = calculate_rankings(trial_list, name_list,
                                  bootstrap_samples, cut=cut)
    plot_ranking(rankings, optimum=optimum, title=title, log=log,
                 save=save, y_min=y_min, y_max=y_max, cut=cut,
                 figsize=figsize, legend_ncols=legend_ncols)


def calculate_rankings(trial_list, name_list, bootstrap_samples=500, cut=50):
    bootstrap_samples = int(bootstrap_samples)
    optimizers = [name[0] for name in name_list]
    pickles = plot_util.load_pickles(name_list, trial_list)
    rankings = dict()

    rs = np.random.RandomState(1)

    combinations = []
    for i in range(bootstrap_samples):
        combination = []
        target = len(optimizers)
        maximum = [len(pickles[name]) for name in optimizers]
        for idx in range(target):
            combination.append(rs.randint(maximum[idx]))
        combinations.append(np.array(combination))

    for optimizer in optimizers:
        rankings[optimizer] = np.zeros((cut+1,), dtype=np.float64)
        rankings[optimizer][0] = np.mean(range(1, len(optimizers) + 1))

    for i in range(1, cut+1):
        num_products = 0

        for combination in combinations:

            ranks = scipy.stats.rankdata(
                [np.round(
                    plot_util.get_best(pickles[optimizers[idx]][number], i), 5)
                 for idx, number in enumerate(combination)])
            num_products += 1
            for j, optimizer in enumerate(optimizers):
                rankings[optimizer][i] += ranks[j]

        for optimizer in optimizers:
            rankings[optimizer][i] = rankings[optimizer][i] / num_products

    return rankings


def plot_ranking(rankings, optimum=0, title="", log=False,
                 save="", y_min=0, y_max=0, figsize=(16, 6), legend_ncols=4,
                 colors=None, markers=None, markersize=6, linewidth=3):
    # check if all optimizers have the same number of runs
    # if np.mean([name[1] for name in name_list]) != name_list[0][1]:
    #    raise Exception("All optimizers must have the same numbers of "
    #                    "experiment runs! %s" % name_list)


    fig = plt.figure(dpi=600, figsize=figsize)
    ax = plt.subplot(111)

    if colors is None:
        colors = plot_util.get_plot_colors()
    if markers is None:
        markers = plot_util.get_empty_iterator()

    for i, optimizer in enumerate(rankings):
        ax.plot(range(0, rankings[optimizer].shape[0]),
                rankings[optimizer],
                marker=markers.next(), markersize=markersize,
                color=colors.next(), linewidth=linewidth,
                label=optimizer.replace("\\", "").
                replace("learned", "d_c").replace("l1", "d_p"))

    ax.legend(loc='upper center',  #bbox_to_anchor=(0.5, -0.05),
              fancybox=True, shadow=True, ncol=legend_ncols,
              labelspacing=0.25, fontsize=12)
    ax.set_xlabel("#Function evaluations")
    ax.set_ylabel("Average rank")
    box = ax.get_position()

    #ax.set_position([box.x0, box.y0 + box.height * 0.1,
    #                 box.width, box.height * 0.9])

    if save != "":
        plt.savefig(save, facecolor='w', edgecolor='w',
                    orientation='portrait', papertype=None, format=None,
                    transparent=False, bbox_inches="tight", pad_inches=0.1)
    else:
        plt.show()
    plt.close(fig)


def get_summed_wins_of_optimizers(trial_list_per_dataset,
                                  name_list_per_dataset,
                                  cut=sys.maxint):
    with open(trial_list_per_dataset[0][0][0]) as fh:
        probing_trial = cPickle.load(fh)
    cut = min(cut, len(probing_trial['trials']))
    # TODO remove this hack!
    cut = 50

    optimizers = []
    for name_list in name_list_per_dataset:
        optimizers.extend([name[0] for name in name_list])
    optimizers = set(optimizers)
    optimizers = list(optimizers)
    optimizers.sort()

    if cut == sys.maxint:
        raise ValueError("You must specify a cut value!")

    summed_wins_of_optimizer = \
        [np.zeros((len(optimizers), len(optimizers))) for i in range(cut + 1)]

    for pkl_list, name_list in itertools.izip(trial_list_per_dataset,
                                              name_list_per_dataset):
        # #######################################################################
        # Statistical stuff for one dataset
        for c in range(1, cut + 1):
            # TODO: this function call can be done in parallel
            wins_of_optimizer = statistics.get_pairwise_wins(pkl_list,
                                                             name_list, cut=c)

            # It can happen that not all optimizers are in all lists,
            # therefore we have to check this first
            for opt1_idx, key in enumerate(optimizers):
                for opt2_idx, key2 in enumerate(optimizers):
                    if key in wins_of_optimizer:
                        tmp_dict = wins_of_optimizer[key]
                        if key2 in tmp_dict:
                            summed_wins_of_optimizer[c][opt1_idx][opt2_idx] += \
                                wins_of_optimizer[key][key2]

    return optimizers, summed_wins_of_optimizer


def plot_summed_wins_of_optimizers(trial_list_per_dataset,
                                   name_list_per_dataset,
                                   save="",  cut=sys.maxint,
                                   figsize=(16, 4), legend_ncols=3,
                                   colors=None, linewidth=3,
                                   markers=None, markersize=6):
    # TODO colors should be a function handle which returns an Iterable!

    # This is a hack
    cut = 50
    optimizers, summed_wins_of_optimizer = get_summed_wins_of_optimizers(
        trial_list_per_dataset, name_list_per_dataset)

    # Make a copy of the colors iterator, because we need it more than once!
    if colors is not None:
        if not isinstance(colors, itertools.cycle):
            raise TypeError()
        else:
            color_values = list()
            for i in range(10):
                r, g, b = colors.next()
                color_values.append((r, g, b))

    if markers is not None:
        if not isinstance(markers, itertools.cycle):
            raise TypeError()
        else:
            marker_values = list()
            for i in range(10):
                marker_values.append(markers.next())

    ################################################################################
    # Plot statistics
    for opt1_idx, key in enumerate(optimizers):
        fig, (ax0, ax1) = plt.subplots(nrows=2, sharex=True, dpi=600,
                                       figsize=figsize)

        if colors is None:
            colors_ = plot_util.get_plot_colors()
        else:
            colors_ = itertools.cycle(color_values)

        if markers is None:
            markers_ = plot_util.get_empty_iterator()
        else:
            markers_ = itertools.cycle(marker_values)

        y_max = 0.

        for opt2_idx, key2 in enumerate(optimizers):
            if opt1_idx == opt2_idx:
                continue

            y = []
            y1 = []
            for i in range(0, cut+1):
                y.append(summed_wins_of_optimizer[i][opt1_idx, opt2_idx]
                         / len(trial_list_per_dataset) * 100)
                y1.append(- summed_wins_of_optimizer[i][opt2_idx, opt1_idx]
                          / len(trial_list_per_dataset) * 100)

            y_max_tmp = max(np.max(y), np.max(np.abs(y1)))
            y_max_tmp = np.ceil(y_max_tmp * 10) / 10.
            y_max = max(y_max_tmp, y_max)

            label = "%s vs %s" % (key, key2)
            label = label.replace("learned", "d_c").replace("l1", "d_p")
            color = colors_.next()
            marker = markers_.next()
            ax0.plot(range(0, cut+1), y, color=color, label=label,
                     linewidth=linewidth, marker=marker, markersize=markersize)
            ax1.plot(range(0, cut+1), y1, color=color, label=label,
                     linewidth=linewidth, marker=marker, markersize=markersize)

        #handles, labels = ax1.get_legend_handles_labels()
        #fig.legend(handles, labels, loc="upper center", fancybox=True,
        #           ncol=legend_ncols, shadow=True)

        ax0.set_xlim((0, cut))
        ax0.set_ylim((0, y_max))
        ax0.set_ylabel("Significant wins (%)")
        ax1.set_xlim((0, cut))
        ax1.set_ylim((-y_max, 0))
        ax1.set_ylabel("Significant losses (%)")
        yticklabels = ax1.get_yticks().tolist()
        #print yticklabels, [item.get_text() for item in yticklabels]
        ax1.set_yticklabels([-int(item) for item in yticklabels])

        ax1.legend(loc="best", fancybox=True, ncol=legend_ncols, shadow=True)

        plt.tight_layout()
        #plt.subplots_adjust(top=0.85)
        plt.xlabel("#Function evaluations")
        if save != "":
            plt.savefig(save % key, facecolor='w', edgecolor='w',
                        orientation='portrait', papertype=None, format=None,
                        transparent=False, bbox_inches="tight", pad_inches=0.1)
        else:
            plt.show()
        plt.close(fig)


# In case we want to display more than the 9 colors present in HPOlib
colors = matplotlib.colors.cnames.copy()
del colors['white']
del colors['whitesmoke']
del colors['ghostwhite']
del colors['aliceblue']
del colors['azure']
del colors['bisque']
del colors['cornsilk']
del colors['floralwhite']
del colors['honeydew']
del colors['ivory']
del colors['lavenderblush']
del colors['lightyellow']
del colors['mintcream']


# Set up a dictionary with keys to all hyperparameter optimization algorithms
def get_all_available_optimizers():
    optimizers = OrderedDict()

    # Helper variables
    better_subset_names = {"all": "All",
                           "pfahringer_2000_experiment1": "Pfahringer",
                           "yogotama_2014": "Yogatama",
                           "bardenet_2013_boost": "Bardenet(Boosting)",
                           "bardenet_2013_nn": "Bardenet(NN)",
                           "subset0": "ForwardFS",
                           "subset1": "EmbeddedFS"}
    num_samples = range(1, 100)
    all_distances = ["random", "l1", "l2", "learned"]
    all_subsets = ["all", "pfahringer_2000_experiment1", "yogotama_2014",
                   "bardenet_2013_boost", "bardenet_2013_nn", "subset0",
                   "subset1"]

    # Add SMBO methods
    optimizers["SMAC"] = "%s/smac_2_06_01-dev_*/smac_2_06_01-dev.pkl"
    optimizers[
        "TPE"] = "%s/hyperopt_august2013_mod_*/hyperopt_august2013_mod.pkl"
    optimizers[
        "Spearmint"] = "%s/spearmint_gitfork_mod_*/spearmint_gitfork_mod.pkl"

    # Random search
    optimizers[
        "random"] = "%s/random_hyperopt_august2013_mod*/random_hyperopt_august2013_mod.pkl"

    for samples in num_samples:
        for dist in all_distances:  # No L2 metric
            for subset in all_subsets:
                if dist == "random" and subset != "all":
                    continue

                optimizers["MI-SMAC(%d,%s,%s)" % (
                    samples, dist, better_subset_names[subset])] = \
                    "%s/bootstrapped" + \
                    "%d_%s_%ssmac_warmstart_*/*smac_warmstart.pkl" % \
                    (samples, dist, subset)

                optimizers["MI-Spearmint(%d,%s,%s)" % (
                    samples, dist, better_subset_names[subset])] = \
                    "%s/" + "bootstrapped%d_%s_%sspearmint_gitfork_mod_" \
                            "*/*spearmint_gitfork_mod.pkl" % (
                                samples, dist, subset)

    return optimizers



import sklearn.ensemble

from autosklearn.pipeline.components.regression.gradient_boosting import \
    GradientBoosting

from .test_base import BaseRegressionComponentTest


class GradientBoostingComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston"] = 0.83773015801058082
    res["boston_n_calls"] = 7
    res["default_boston_iterative"] = 0.83773015801058082
    res["default_boston_sparse"] = None
    res["default_boston_iterative_sparse"] = None
    res["default_diabetes"] = 0.33574377250653153
    res["diabetes_n_calls"] = 7
    res["default_diabetes_iterative"] = 0.33574377250653153
    res["default_diabetes_sparse"] = None
    res["default_diabetes_iterative_sparse"] = None

    sk_mod = sklearn.ensemble.GradientBoostingRegressor
    module = GradientBoosting
    step_hyperparameter = {
        'name': 'n_estimators',
        'value': 100,
    }


import os
import logging
import shutil
import sys
import time
import unittest
import unittest.mock

import numpy as np

this_directory = os.path.dirname(__file__)
sys.path.append(this_directory)

from autosklearn.ensemble_builder import EnsembleBuilder, Y_ENSEMBLE, Y_VALID, Y_TEST
from autosklearn.metrics import roc_auc


class BackendMock(object):
    
    def __init__(self):
        this_directory = os.path.abspath(
            os.path.dirname(__file__)
        )
        self.temporary_directory = os.path.join(
            this_directory, 'data',
        )
    
    def load_targets_ensemble(self):
        with open(os.path.join(
            self.temporary_directory,
            ".auto-sklearn",
            "predictions_ensemble",
            "predictions_ensemble_true.npy"
        ),"rb") as fp:
            y = np.load(fp)
        return y
    
class EnsembleBuilderMemMock(EnsembleBuilder):
    
    def fit_ensemble(self,selected_keys):
        np.ones([10000000,1000000])


class EnsembleTest(unittest.TestCase):
    def setUp(self):
        self.backend = BackendMock()

    def tearDown(self):
        pass

    def testRead(self):
        
        ensbuilder = EnsembleBuilder(
            backend=self.backend,
            dataset_name="TEST",
            task_type=1,  #Binary Classification
            metric=roc_auc,
            limit=-1,  # not used,
            seed=0,  # important to find the test files
        )

        success = ensbuilder.read_ensemble_preds()
        self.assertTrue(success, str(ensbuilder.read_preds))
        self.assertEqual(len(ensbuilder.read_preds), 2)

        filename = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_1.npy"
        )
        self.assertEqual(ensbuilder.read_preds[filename]["ens_score"], 0.5)

        filename = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_2.npy"
        )
        self.assertEqual(ensbuilder.read_preds[filename]["ens_score"], 1.0)
                    
    def testNBest(self):
        
        ensbuilder = EnsembleBuilder(
            backend=self.backend,
            dataset_name="TEST",
            task_type=1,  #Binary Classification
            metric=roc_auc,
            limit=-1, # not used,
            seed=0, # important to find the test files
            ensemble_nbest=1,
        )
        
        ensbuilder.read_ensemble_preds()
        sel_keys = ensbuilder.get_n_best_preds()

        self.assertEquals(len(sel_keys), 1)

        fixture = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_2.npy"
        )
        self.assertEquals(sel_keys[0], fixture)
        
    def testFallBackNBest(self):
        
        ensbuilder = EnsembleBuilder(backend=self.backend, 
                                    dataset_name="TEST",
                                    task_type=1,  #Binary Classification
                                    metric=roc_auc,
                                    limit=-1, # not used,
                                    seed=0, # important to find the test files
                                    ensemble_nbest=1
                                    )
        
        ensbuilder.read_ensemble_preds()

        filename = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_2.npy"
        )
        ensbuilder.read_preds[filename]["ens_score"] = -1

        filename = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_1.npy"
        )
        ensbuilder.read_preds[filename]["ens_score"] = -1
        
        sel_keys = ensbuilder.get_n_best_preds()

        fixture = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_1.npy"
        )
        self.assertEquals(sel_keys[0], fixture)
        
    def testGetValidTestPreds(self):
        
        ensbuilder = EnsembleBuilder(backend=self.backend, 
                                    dataset_name="TEST",
                                    task_type=1,  #Binary Classification
                                    metric=roc_auc,
                                    limit=-1, # not used,
                                    seed=0, # important to find the test files
                                    ensemble_nbest=1
                                    )
        
        ensbuilder.read_ensemble_preds()
        
        d2 = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_2.npy"
        )
        d1 = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_1.npy"
        )
        
        sel_keys = ensbuilder.get_n_best_preds()
        
        ensbuilder.get_valid_test_preds(selected_keys=sel_keys)
        
        # selected --> read valid and test predictions
        self.assertIsNotNone(ensbuilder.read_preds[d2][Y_VALID])
        self.assertIsNotNone(ensbuilder.read_preds[d2][Y_TEST])
        
        # not selected --> should still be None
        self.assertIsNone(ensbuilder.read_preds[d1][Y_VALID])
        self.assertIsNone(ensbuilder.read_preds[d1][Y_TEST])
        
    def testEntireEnsembleBuilder(self):
        
        ensbuilder = EnsembleBuilder(
            backend=self.backend,
            dataset_name="TEST",
            task_type=1,  #Binary Classification
            metric=roc_auc,
            limit=-1, # not used,
            seed=0, # important to find the test files
            ensemble_nbest=2,
        )
        ensbuilder.SAVE2DISC = False
        
        ensbuilder.read_ensemble_preds()

        d2 = os.path.join(
            self.backend.temporary_directory,
            ".auto-sklearn/predictions_ensemble/predictions_ensemble_0_2.npy"
        )

        sel_keys = ensbuilder.get_n_best_preds()
        self.assertGreater(len(sel_keys), 0)
        
        ensemble = ensbuilder.fit_ensemble(selected_keys=sel_keys)
        print(ensemble, sel_keys)
        
        n_sel_valid, n_sel_test = ensbuilder.get_valid_test_preds(selected_keys=sel_keys)
        
        # both valid and test prediction files are available
        self.assertGreater(len(n_sel_valid), 0)
        self.assertEqual(n_sel_valid, n_sel_test)

        y_valid = ensbuilder.predict(
            set_="valid",
            ensemble=ensemble,
            selected_keys=n_sel_valid,
            n_preds=len(sel_keys),
            index_run=1,
        )
        y_test = ensbuilder.predict(
            set_="test",
            ensemble=ensemble,
            selected_keys=n_sel_test,
            n_preds=len(sel_keys),
            index_run=1,
        )

        # predictions for valid and test are the same
        # --> should results in the same predictions
        np.testing.assert_array_almost_equal(y_valid, y_test)

        # since d2 provides perfect predictions
        # it should get a higher weight
        # so that y_valid should be exactly y_valid_d2
        y_valid_d2 = ensbuilder.read_preds[d2][Y_VALID][:, 1]
        np.testing.assert_array_almost_equal(y_valid, y_valid_d2)
        
    def testMain(self):
        
        ensbuilder = EnsembleBuilder(backend=self.backend, 
                                    dataset_name="TEST",
                                    task_type=1,  #Binary Classification
                                    metric=roc_auc,
                                    limit=-1, # not used,
                                    seed=0, # important to find the test files
                                    ensemble_nbest=2,
                                    max_iterations=1 # prevents infinite loop
                                    )
        ensbuilder.SAVE2DISC = False
        
        ensbuilder.main()
        
        self.assertEqual(len(ensbuilder.read_preds), 2)
        self.assertIsNotNone(ensbuilder.last_hash)
        self.assertIsNotNone(ensbuilder.y_true_ensemble)
        
    def testLimit(self):
        
                
        ensbuilder = EnsembleBuilderMemMock(backend=self.backend, 
                                            dataset_name="TEST",
                                            task_type=1,  #Binary Classification
                                            metric=roc_auc,
                                            limit=1000, # not used,
                                            seed=0, # important to find the test files
                                            ensemble_nbest=10,
                                            max_iterations=1, # prevents infinite loop
                                            memory_limit=10 # small memory limit to trigger MemoryException
                                            )
        ensbuilder.SAVE2DISC = False
        
        ensbuilder.run()
        
        # it should try to reduce ensemble_nbest until it also failed at 2
        self.assertEqual(ensbuilder.ensemble_nbest,1)



import sklearn.ensemble

from autosklearn.pipeline.components.classification.gradient_boosting import \
    GradientBoostingClassifier

from .test_base import BaseClassificationComponentTest


class GradientBoostingComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.92
    res["iris_n_calls"] = 7
    res["default_iris_iterative"] = 0.92
    res["default_iris_proba"] = 0.48109031836615801
    res["default_iris_sparse"] = -1
    res["default_digits"] = 0.80206435944140864
    res["digits_n_calls"] = 7
    res["default_digits_iterative"] = 0.80206435944140864
    res["default_digits_binary"] = 0.98178506375227692
    res["default_digits_multilabel"] = -1
    res["default_digits_multilabel_proba"] = -1

    sk_mod = sklearn.ensemble.ExtraTreesClassifier
    module = GradientBoostingClassifier
    step_hyperparameter = {
        'name': 'n_estimators',
        'value': 100,
    }


import unittest

import numpy as np
import scipy.sparse
from sklearn.utils.testing import assert_array_almost_equal
import sklearn.tree
import sklearn.datasets
import sklearn.model_selection
import sklearn.pipeline
import openml

from autosklearn.pipeline.implementations.OneHotEncoder import OneHotEncoder

dense1 = np.array([[0, 1, 0],
                   [0, 0, 0],
                   [1, 1, 0]])
dense1_1h = np.array([[1, 0, 0, 1, 1],
                     [1, 0, 1, 0, 1],
                     [0, 1, 0, 1, 1]])
dense1_1h_minimum_fraction = np.array([[0, 1, 0, 1, 1],
                                       [0, 1, 1, 0, 1],
                                       [1, 0, 0, 1, 1]])

# Including NaNs
dense2 = np.array([[0, np.NaN, 0],
                   [np.NaN, 0, 2],
                   [1, 1, 1],
                   [np.NaN, 0, 1]])
dense2_1h = np.array([[0, 1, 0, 1, 0, 0, 1, 0, 0],
                      [1, 0, 0, 0, 1, 0, 0, 0, 1],
                      [0, 0, 1, 0, 0, 1, 0, 1, 0],
                      [1, 0, 0, 0, 1, 0, 0, 1, 0]])

dense2_1h_minimum_fraction = np.array([[1, 0, 1, 0, 1, 0],
                                       [0, 1, 0, 1, 1, 0],
                                       [1, 0, 1, 0, 0, 1],
                                       [0, 1, 0, 1, 0, 1]])

dense2_partial_1h = np.array([[0., 1., 0., 1., 0., 0., 0.],
                              [1., 0., 0., 0., 1., 0., 2.],
                              [0., 0., 1., 0., 0., 1., 1.],
                              [1., 0., 0., 0., 1., 0., 1.]])

dense2_1h_minimum_fraction_as_sparse = np.array([[0, 0, 1, 0, 0, 0],
                                                 [0, 1, 0, 0, 1, 0],
                                                 [1, 0, 0, 1, 0, 1],
                                                 [0, 1, 0, 0, 0, 1]])

# All NaN slice
dense3 = np.array([[0, 1, np.NaN],
                   [1, 0, np.NaN]])
dense3_1h = np.array([[1, 0, 0, 1, 1],
                      [0, 1, 1, 0, 1]])

sparse1 = scipy.sparse.csc_matrix(([3, 2, 1, 1, 2, 3],
                                   ((1, 4, 5, 2, 3, 5),
                                    (0, 0, 0, 1, 1, 1))), shape=(6, 2))
sparse1_1h = scipy.sparse.csc_matrix(([1, 1, 1, 1, 1, 1],
                                      ((5, 4, 1, 2, 3, 5),
                                       (0, 1, 2, 3, 4, 5))), shape=(6, 6))
sparse1_paratial_1h = scipy.sparse.csc_matrix(([1, 1, 1, 1, 2, 3],
                                               ((5, 4, 1, 2, 3, 5),
                                                (0, 1, 2, 3, 3, 3))),
                                              shape=(6, 4))

# All zeros slice
sparse2 = scipy.sparse.csc_matrix(([2, 1, 0, 0, 0, 0],
                                   ((1, 4, 5, 2, 3, 5),
                                    (0, 0, 0, 1, 1, 1))), shape=(6, 2))
sparse2_1h = scipy.sparse.csc_matrix(([1, 1, 1, 1, 1, 1],
                                      ((5, 4, 1, 2, 3, 5),
                                       (0, 1, 2, 3, 3, 3))), shape=(6, 4))

sparse2_csr = scipy.sparse.csr_matrix(([2, 1, 0, 0, 0, 0],
                                      ((1, 4, 5, 2, 3, 5),
                                       (0, 0, 0, 1, 1, 1))), shape=(6, 2))
sparse2_csr_1h = scipy.sparse.csr_matrix(([1, 1, 1, 1, 1, 1],
                                         ((5, 4, 1, 2, 3, 5),
                                          (0, 1, 2, 3, 3, 3))), shape=(6, 4))


class TestOneHotEncoder(unittest.TestCase):
    def test_dense1(self):
        self._fit_then_transform(dense1_1h, dense1)
        self._fit_then_transform_dense(dense1_1h, dense1)

    def test_dense1_minimum_fraction(self):
        self._fit_then_transform(dense1_1h_minimum_fraction, dense1, minimum_fraction=0.5)
        self._fit_then_transform_dense(dense1_1h_minimum_fraction, dense1, minimum_fraction=0.5)

    def test_dense2(self):
        self._fit_then_transform(dense2_1h, dense2)
        self._fit_then_transform_dense(dense2_1h, dense2)

    def test_dense2_minimum_fraction(self):
        self._fit_then_transform(dense2_1h_minimum_fraction, dense2,
                                 minimum_fraction=0.3)
        self._fit_then_transform_dense(dense2_1h_minimum_fraction, dense2,
                                       minimum_fraction=0.3)

    def test_dense2_with_non_sparse_components(self):
        self._fit_then_transform(dense2_partial_1h, dense2,
                                 categorical_features=[True, True, False])
        self._fit_then_transform_dense(dense2_partial_1h, dense2,
                                       categorical_features=[True, True, False])

    # Minimum fraction is not too interesting here...
    def test_dense3(self):
        self._fit_then_transform(dense3_1h, dense3)
        self._fit_then_transform_dense(dense3_1h, dense3)

    def test_sparse1(self):
        self._fit_then_transform(sparse1_1h.todense(), sparse1)
        self._fit_then_transform_dense(sparse1_1h.todense(), sparse1)

    def test_sparse1_minimum_fraction(self):
        expected = np.array([[0, 1, 0, 0, 1, 1],
                             [0, 0, 1, 1, 0, 1]], dtype=float).transpose()
        self._fit_then_transform(expected, sparse1,
                                 minimum_fraction=0.5)
        self._fit_then_transform_dense(expected, sparse1,
                                       minimum_fraction=0.5)

    def test_sparse1_with_non_sparse_components(self):
        self._fit_then_transform(sparse1_paratial_1h.todense(), sparse1,
                                 categorical_features=[True, False])
        # This test does not apply here. The sparse matrix will be cut into a
        #  continouos and a categorical part, after one hot encoding only the
        #  categorical part is an array, the continuous part will still be a
        # sparse matrix. Therefore, the OHE will only return a sparse matrix
        #self._fit_then_transform_dense(sparse1_paratial_1h.todense(), sparse1,
        #                              categorical_features=[True, False])

    def test_sparse2(self):
        self._fit_then_transform(sparse2_1h.todense(), sparse2)
        self._fit_then_transform_dense(sparse2_1h.todense(), sparse2)

    def test_sparse2_minimum_fraction(self):
        expected = np.array([[0, 1, 0, 0, 1, 1],
                             [0, 0, 1, 1, 0, 1]], dtype=float).transpose()
        self._fit_then_transform(expected, sparse2,
                                 minimum_fraction=0.5)
        self._fit_then_transform_dense(expected, sparse2,
                                       minimum_fraction=0.5)

    def test_sparse2_csr(self):
        self._fit_then_transform(sparse2_csr_1h.todense(), sparse2_csr)
        self._fit_then_transform_dense(sparse2_csr_1h.todense(), sparse2_csr)

    def test_sparse_on_dense2_minimum_fraction(self):
        sparse = scipy.sparse.csr_matrix(dense2)
        self._fit_then_transform(dense2_1h_minimum_fraction_as_sparse, sparse,
                                 minimum_fraction=0.5)
        self._fit_then_transform_dense(dense2_1h_minimum_fraction_as_sparse, sparse,
                                       minimum_fraction=0.5)

    def _fit_then_transform(self, expected, input, categorical_features='all',
                            minimum_fraction=None):
        # Test fit_transform
        input_copy = input.copy()
        ohe = OneHotEncoder(categorical_features=categorical_features,
                            minimum_fraction=minimum_fraction)
        transformation = ohe.fit_transform(input)
        self.assertIsInstance(transformation, scipy.sparse.csr_matrix)
        assert_array_almost_equal(expected.astype(float),
                                  transformation.todense())
        self._check_arrays_equal(input, input_copy)

        # Test fit, and afterwards transform
        ohe2 = OneHotEncoder(categorical_features=categorical_features,
                             minimum_fraction=minimum_fraction)
        ohe2.fit(input)
        transformation = ohe2.transform(input)
        self.assertIsInstance(transformation, scipy.sparse.csr_matrix)
        assert_array_almost_equal(expected, transformation.todense())
        self._check_arrays_equal(input, input_copy)

    def _fit_then_transform_dense(self, expected, input,
                                  categorical_features='all',
                                  minimum_fraction=None):
        input_copy = input.copy()
        ohe = OneHotEncoder(categorical_features=categorical_features,
                            sparse=False, minimum_fraction=minimum_fraction)
        transformation = ohe.fit_transform(input)
        self.assertIsInstance(transformation, np.ndarray)
        assert_array_almost_equal(expected, transformation)
        self._check_arrays_equal(input, input_copy)

        ohe2 = OneHotEncoder(categorical_features=categorical_features,
                             sparse=False, minimum_fraction=minimum_fraction)
        ohe2.fit(input)
        transformation = ohe2.transform(input)
        self.assertIsInstance(transformation, np.ndarray)
        assert_array_almost_equal(expected, transformation)
        self._check_arrays_equal(input, input_copy)

    def _check_arrays_equal(self, a1, a2):
        if scipy.sparse.issparse(a1):
            a1 = a1.toarray()
        if scipy.sparse.issparse(a2):
            a2 = a2.toarray()
        assert_array_almost_equal(a1, a2)

    def test_transform_with_unknown_value(self):
        input = np.array(((0, 1, 2, 3, 4, 5), (0, 1, 2, 3, 4, 5))).transpose()
        ohe = OneHotEncoder()
        ohe.fit(input)
        test_data = np.array(((0, 1, 2, 6), (0, 1, 6, 7))).transpose()
        output = ohe.transform(test_data).todense()
        self.assertEqual(5, np.sum(output))

        input = np.array(((0, 1, 2, 3, 4, 5), (0, 1, 2, 3, 4, 5))).transpose()
        ips = scipy.sparse.csr_matrix(input)
        ohe = OneHotEncoder()
        ohe.fit(ips)
        test_data = np.array(((0, 1, 2, 6), (0, 1, 6, 7))).transpose()
        tds = scipy.sparse.csr_matrix(test_data)
        output = ohe.transform(tds).todense()
        self.assertEqual(3, np.sum(output))

    def test_classification_workflow(self):
        task = openml.tasks.get_task(254)
        X, y = task.get_X_and_y()

        ohe = OneHotEncoder(categorical_features=[True]*22)
        tree = sklearn.tree.DecisionTreeClassifier(random_state=1)
        pipeline = sklearn.pipeline.Pipeline((('ohe', ohe), ('tree', tree)))

        X_train, X_test, y_train, y_test = \
            sklearn.model_selection.train_test_split(X, y, random_state=3,
                                                     train_size=0.5,
                                                     test_size=0.5)
        pipeline.fit(X_train, y_train)
        self.assertEqual(np.mean(y_train == pipeline.predict(X_train)), 1)
        # With an incorrect copy operation the OneHotEncoder would rearrange
        # the data in such a way that the accuracy would drop to 66%
        self.assertEqual(np.mean(y_test == pipeline.predict(X_test)), 1)
